r"""
Discrete valuations on function fields

AUTHORS:

- Julian Rüth (2016-10-16): initial version

EXAMPLES:

We can create classical valuations that correspond to finite and infinite
places on a rational function field::

    sage: K.<x> = FunctionField(QQ)
    sage: v = K.valuation(1); v
    (x - 1)-adic valuation
    sage: v = K.valuation(x^2 + 1); v
    (x^2 + 1)-adic valuation
    sage: v = K.valuation(1/x); v
    Valuation at the infinite place

Note that we can also specify valuations which do not correspond to a place of
the function field::

    sage: R.<x> = QQ[]
    sage: w = valuations.GaussValuation(R, QQ.valuation(2))
    sage: v = K.valuation(w); v
    2-adic valuation

Valuations on a rational function field can then be extended to finite
extensions::

    sage: v = K.valuation(x - 1); v
    (x - 1)-adic valuation
    sage: R.<y> = K[]
    sage: L.<y> = K.extension(y^2 - x)                                                  # needs sage.rings.function_field
    sage: w = v.extensions(L); w                                                        # needs sage.rings.function_field
    [[ (x - 1)-adic valuation, v(y + 1) = 1 ]-adic valuation,
     [ (x - 1)-adic valuation, v(y - 1) = 1 ]-adic valuation]

TESTS:

Run test suite for classical places over rational function fields::

    sage: K.<x> = FunctionField(QQ)
    sage: v = K.valuation(1)
    sage: TestSuite(v).run(max_runs=100) # long time

    sage: v = K.valuation(x^2 + 1)
    sage: TestSuite(v).run(max_runs=100) # long time

    sage: v = K.valuation(1/x)
    sage: TestSuite(v).run(max_runs=100) # long time

Run test suite over classical places of finite extensions::

    sage: K.<x> = FunctionField(QQ)
    sage: v = K.valuation(x - 1)
    sage: R.<y> = K[]
    sage: L.<y> = K.extension(y^2 - x)                                                  # needs sage.rings.function_field
    sage: ws = v.extensions(L)                                                          # needs sage.rings.function_field
    sage: for w in ws: TestSuite(w).run(max_runs=100)   # long time                     # needs sage.rings.function_field

Run test suite for valuations that do not correspond to a classical place::

    sage: K.<x> = FunctionField(QQ)
    sage: R.<x> = QQ[]
    sage: v = GaussValuation(R, QQ.valuation(2))
    sage: w = K.valuation(v)
    sage: TestSuite(w).run() # long time

Run test suite for a non-classical valuation that does not correspond to an
affinoid contained in the unit disk::

    sage: w = K.valuation((w, K.hom(K.gen()/2), K.hom(2*K.gen()))); w
    2-adic valuation (in Rational function field in x over Rational Field after x |--> 1/2*x)
    sage: TestSuite(w).run() # long time

Run test suite for some other classical places over large ground fields::

    sage: K.<t> = FunctionField(GF(3))
    sage: M.<x> = FunctionField(K)
    sage: v = M.valuation(x^3 - t)
    sage: TestSuite(v).run(max_runs=10)         # long time

Run test suite for extensions over the infinite place::

    sage: K.<x> = FunctionField(QQ)
    sage: v = K.valuation(1/x)
    sage: R.<y> = K[]
    sage: L.<y> = K.extension(y^2 - 1/(x^2 + 1))                                        # needs sage.rings.function_field
    sage: w = v.extensions(L)                                                           # needs sage.rings.function_field
    sage: TestSuite(w).run()                    # long time                             # needs sage.rings.function_field

Run test suite for a valuation with `v(1/x) > 0` which does not come from a
classical valuation of the infinite place::

    sage: K.<x> = FunctionField(QQ)
    sage: R.<x> = QQ[]
    sage: w = GaussValuation(R, QQ.valuation(2)).augmentation(x, 1)
    sage: w = K.valuation(w)
    sage: v = K.valuation((w, K.hom([~K.gen()]), K.hom([~K.gen()])))
    sage: TestSuite(v).run() # long time

Run test suite for extensions which come from the splitting in the base field::

    sage: K.<x> = FunctionField(QQ)
    sage: v = K.valuation(x^2 + 1)
    sage: L.<x> = FunctionField(GaussianIntegers().fraction_field())
    sage: ws = v.extensions(L)                                                          # needs sage.rings.function_field
    sage: for w in ws: TestSuite(w).run(max_runs=100)   # long time                     # needs sage.rings.function_field

Run test suite for a finite place with residual degree and ramification::

    sage: K.<t> = FunctionField(GF(3))
    sage: L.<x> = FunctionField(K)
    sage: v = L.valuation(x^6 - t)
    sage: TestSuite(v).run(max_runs=10) # long time

Run test suite for a valuation which is backed by limit valuation::

    sage: K.<x> = FunctionField(QQ)
    sage: R.<y> = K[]
    sage: L.<y> = K.extension(y^2 - (x^2 + x + 1))
    sage: v = K.valuation(x - 1)
    sage: w = v.extension(L)                                                            # needs sage.rings.function_field
    sage: TestSuite(w).run()                    # long time                             # needs sage.rings.function_field

Run test suite for a valuation which sends an element to `-\infty`::

    sage: R.<x> = QQ[]
    sage: v = GaussValuation(QQ['x'], QQ.valuation(2)).augmentation(x, infinity)
    sage: K.<x> = FunctionField(QQ)
    sage: w = K.valuation(v)
    sage: TestSuite(w).run() # long time

REFERENCES:

An overview of some computational tools relating to valuations on function
fields can be found in Section 4.6 of [Rüt2014]_. Most of this was originally
developed for number fields in [Mac1936I]_ and [Mac1936II]_.
"""

# ****************************************************************************
#       Copyright (C) 2016-2018 Julian Rüth <julian.rueth@fsfe.org>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.rings.rational_field import QQ
from sage.rings.valuation.mapped_valuation import (
    FiniteExtensionFromLimitValuation,
    MappedValuation_base,
)
from sage.rings.valuation.trivial_valuation import TrivialValuation
from sage.rings.valuation.valuation import (
    DiscretePseudoValuation,
    DiscreteValuation,
    InfiniteDiscretePseudoValuation,
    NegativeInfiniteDiscretePseudoValuation,
)
from sage.structure.factory import UniqueFactory


class FunctionFieldValuationFactory(UniqueFactory):
    r"""
    Create a valuation on ``domain`` corresponding to ``prime``.

    INPUT:

    - ``domain`` -- a function field

    - ``prime`` -- a place of the function field, a valuation on a subring, or
      a valuation on another function field together with information for
      isomorphisms to and from that function field

    EXAMPLES::

        sage: K.<x> = FunctionField(QQ)
        sage: v = K.valuation(1); v  # indirect doctest
        (x - 1)-adic valuation
        sage: v(x)
        0
        sage: v(x - 1)
        1

    See :meth:`sage.rings.function_field.function_field.FunctionField.valuation` for further examples.
    """
    def create_key_and_extra_args(self, domain, prime):
        r"""
        Create a unique key which identifies the valuation given by ``prime``
        on ``domain``.

        TESTS:

        We specify a valuation on a function field by two different means and
        get the same object::

            sage: K.<x> = FunctionField(QQ)
            sage: v = K.valuation(x - 1)  # indirect doctest

            sage: R.<x> = QQ[]
            sage: w = GaussValuation(R, valuations.TrivialValuation(QQ)).augmentation(x - 1, 1)
            sage: K.valuation(w) is v
            True

        The normalization is, however, not smart enough, to unwrap
        substitutions that turn out to be trivial::

            sage: w = GaussValuation(R, QQ.valuation(2))
            sage: w = K.valuation(w)
            sage: w is K.valuation((w, K.hom([~K.gen()]), K.hom([~K.gen()])))
            False
        """
        from sage.categories.function_fields import FunctionFields
        if domain not in FunctionFields():
            raise ValueError("Domain must be a function field.")

        if isinstance(prime, tuple):
            if len(prime) == 3:
                # prime is a triple of a valuation on another function field with
                # isomorphism information
                return self.create_key_and_extra_args_from_valuation_on_isomorphic_field(domain, prime[0], prime[1], prime[2])

        from sage.rings.valuation.valuation_space import DiscretePseudoValuationSpace
        if prime.parent() is DiscretePseudoValuationSpace(domain):
            # prime is already a valuation of the requested domain
            # if we returned (domain, prime), we would break caching
            # because this element has been created from a different key
            # Instead, we return the key that was used to create prime
            # so the caller gets back a correctly cached version of prime
            if not hasattr(prime, "_factory_data"):
                raise NotImplementedError("Valuations on function fields must be unique and come out of the FunctionFieldValuation factory but %r has been created by other means" % (prime,))
            return prime._factory_data[2], {}

        if prime in domain:
            # prime defines a place
            return self.create_key_and_extra_args_from_place(domain, prime)
        if prime.parent() is DiscretePseudoValuationSpace(domain._ring):
            # prime is a discrete (pseudo-)valuation on the polynomial ring
            # that the domain is constructed from
            return self.create_key_and_extra_args_from_valuation(domain, prime)
        if domain.base_field() is not domain:
            # prime might define a valuation on a subring of domain and have a
            # unique extension to domain
            base_valuation = domain.base_field().valuation(prime)
            return self.create_key_and_extra_args_from_valuation(domain, base_valuation)
        from sage.rings.ideal import Ideal_generic
        if isinstance(prime, Ideal_generic):
            raise NotImplementedError("a place cannot be given by an ideal yet")

        raise NotImplementedError("argument must be a place or a pseudo-valuation on a supported subring but %r does not satisfy this for the domain %r" % (prime, domain))

    def create_key_and_extra_args_from_place(self, domain, generator):
        r"""
        Create a unique key which identifies the valuation at the place
        specified by ``generator``.

        TESTS:

            sage: K.<x> = FunctionField(QQ)
            sage: v = K.valuation(1/x)  # indirect doctest
        """
        if generator not in domain.base_field():
            raise NotImplementedError("a place must be defined over a rational function field")

        if domain.base_field() is not domain:
            # if this is an extension field, construct the unique place over
            # the place on the subfield
            return self.create_key_and_extra_args(domain, domain.base_field().valuation(generator))

        if generator in domain.constant_base_field():
            # generator is a constant, we associate to it the place which
            # corresponds to the polynomial (x - generator)
            return self.create_key_and_extra_args(domain, domain.gen() - generator)

        if generator in domain._ring:
            # generator is a polynomial
            generator = domain._ring(generator)
            if not generator.is_monic():
                raise ValueError("place must be defined by a monic polynomial but %r is not monic" % (generator,))
            if not generator.is_irreducible():
                raise ValueError("place must be defined by an irreducible polynomial but %r factors over %r" % (generator, domain._ring))
            # we construct the corresponding valuation on the polynomial ring
            # with v(generator) = 1
            from sage.rings.valuation.gauss_valuation import GaussValuation
            valuation = GaussValuation(domain._ring, TrivialValuation(domain.constant_base_field())).augmentation(generator, 1)
            return self.create_key_and_extra_args(domain, valuation)
        if generator == ~domain.gen():
            # generator is 1/x, the infinite place
            return (domain, (domain.valuation(domain.gen()), domain.hom(~domain.gen()), domain.hom(~domain.gen()))), {}
        raise ValueError("a place must be given by an irreducible polynomial or the inverse of the generator; %r does not define a place over %r" % (generator, domain))

    def create_key_and_extra_args_from_valuation(self, domain, valuation):
        r"""
        Create a unique key which identifies the valuation which extends
        ``valuation``.

        TESTS:

            sage: K.<x> = FunctionField(QQ)
            sage: R.<x> = QQ[]
            sage: w = GaussValuation(R, valuations.TrivialValuation(QQ)).augmentation(x - 1, 1)
            sage: v = K.valuation(w) # indirect doctest

        Check that :issue:`25294` has been resolved::

            sage: K.<x> = FunctionField(QQ)
            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y^3 + 1/x^3*y + 2/x^4)                            # needs sage.rings.function_field
            sage: v = K.valuation(x)                                                    # needs sage.rings.function_field
            sage: v.extensions(L)                                                       # needs sage.rings.function_field
            [[ (x)-adic valuation, v(y) = 1 ]-adic valuation
               (in Function field in y defined by y^3 + x*y + 2*x^2 after y |--> 1/x^2*y),
             [ (x)-adic valuation, v(y) = 1/2 ]-adic valuation
               (in Function field in y defined by y^3 + x*y + 2*x^2 after y |--> 1/x^2*y)]
        """
        # this should have been handled by create_key already
        assert valuation.domain() is not domain

        if valuation.domain() is domain._ring:
            if domain.base_field() is not domain:
                vK = valuation.restriction(valuation.domain().base_ring())
                if vK.domain() is not domain.base_field():
                    raise ValueError("valuation must extend a valuation on the base field but %r extends %r whose domain is not %r" % (valuation, vK, domain.base_field()))
                # Valuation is an approximant that describes a single valuation
                # on domain.
                # For uniqueness of valuations (which provides better caching
                # and easier pickling) we need to find a normal form of
                # valuation, i.e., the smallest approximant that describes this
                # valuation
                approximants = vK.mac_lane_approximants(domain.polynomial(), require_incomparability=True)
                approximant = vK.mac_lane_approximant(domain.polynomial(), valuation, approximants)
                return (domain, approximant), {'approximants': approximants}
            # on a rational function field K(x), any valuation on K[x] that
            # does not have an element with valuation -infty extends to a
            # pseudo-valuation on K(x)
            if valuation.is_negative_pseudo_valuation():
                raise ValueError("there must not be an element of valuation -Infinity in the domain of valuation %r" % (valuation,))
            return (domain, valuation), {}

        if valuation.domain().is_subring(domain.base_field()):
            # valuation is defined on a subring of this function field, try to lift it
            return self.create_key_and_extra_args(domain, valuation.extension(domain))

        raise NotImplementedError("extension of valuation from %r to %r not implemented yet" % (valuation.domain(), domain))

    def create_key_and_extra_args_from_valuation_on_isomorphic_field(self, domain, valuation, to_valuation_domain, from_valuation_domain):
        r"""
        Create a unique key which identifies the valuation which is
        ``valuation`` after mapping through ``to_valuation_domain``.

        TESTS::

            sage: K.<x> = FunctionField(GF(2))
            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y^2 + y + x^3)                                    # needs sage.rings.function_field
            sage: v = K.valuation(1/x)                                                  # needs sage.rings.function_field
            sage: w = v.extension(L)  # indirect doctest                                # needs sage.rings.function_field
        """
        from sage.categories.function_fields import FunctionFields
        if valuation.domain() not in FunctionFields():
            raise ValueError("valuation must be defined over an isomorphic function field but %r is not a function field" % (valuation.domain(),))

        from sage.categories.homset import Hom
        if to_valuation_domain not in Hom(domain, valuation.domain()):
            raise ValueError("to_valuation_domain must map from %r to %r but %r maps from %r to %r" % (domain, valuation.domain(), to_valuation_domain, to_valuation_domain.domain(), to_valuation_domain.codomain()))
        if from_valuation_domain not in Hom(valuation.domain(), domain):
            raise ValueError("from_valuation_domain must map from %r to %r but %r maps from %r to %r" % (valuation.domain(), domain, from_valuation_domain, from_valuation_domain.domain(), from_valuation_domain.codomain()))

        if domain is domain.base():
            if valuation.domain() is not valuation.domain().base() or valuation.domain().constant_base_field() != domain.constant_base_field():
                raise NotImplementedError("maps must be isomorphisms with a rational function field over the same base field, not with %r" % (valuation.domain(),))
            if domain != valuation.domain():
                # make it harder to create different representations of the same valuation
                # (nothing bad happens if we did, but >= and <= are only implemented when this is the case.)
                raise NotImplementedError("domain and valuation.domain() must be the same rational function field but %r is not %r" % (domain, valuation.domain()))
        else:
            if domain.base() is not valuation.domain().base():
                raise NotImplementedError("domain and valuation.domain() must have the same base field but %r is not %r" % (domain.base(), valuation.domain().base()))
            if to_valuation_domain != domain.hom([to_valuation_domain(domain.gen())]):
                raise NotImplementedError("to_valuation_domain must be trivial on the base fields but %r is not %r" % (to_valuation_domain, domain.hom([to_valuation_domain(domain.gen())])))
            if from_valuation_domain != valuation.domain().hom([from_valuation_domain(valuation.domain().gen())]):
                raise NotImplementedError("from_valuation_domain must be trivial on the base fields but %r is not %r" % (from_valuation_domain, valuation.domain().hom([from_valuation_domain(valuation.domain().gen())])))
            if to_valuation_domain(domain.gen()) == valuation.domain().gen():
                raise NotImplementedError("to_valuation_domain seems to be trivial but trivial maps would currently break partial orders of valuations")

        if from_valuation_domain(to_valuation_domain(domain.gen())) != domain.gen():
            # only a necessary condition
            raise ValueError("to_valuation_domain and from_valuation_domain are not inverses of each other")

        return (domain, (valuation, to_valuation_domain, from_valuation_domain)), {}

    def create_object(self, version, key, **extra_args):
        r"""
        Create the valuation specified by ``key``.

        EXAMPLES::

            sage: K.<x> = FunctionField(QQ)
            sage: R.<x> = QQ[]
            sage: w = valuations.GaussValuation(R, QQ.valuation(2))
            sage: v = K.valuation(w); v  # indirect doctest
            2-adic valuation
        """
        domain, valuation = key
        from sage.rings.valuation.valuation_space import DiscretePseudoValuationSpace
        parent = DiscretePseudoValuationSpace(domain)

        if isinstance(valuation, tuple) and len(valuation) == 3:
            valuation, to_valuation_domain, from_valuation_domain = valuation
            if domain is domain.base() and valuation.domain() is valuation.domain().base():
                if valuation == valuation.domain().valuation(valuation.domain().gen()):
                    if to_valuation_domain != domain.hom([~valuation.domain().gen()]) or from_valuation_domain != valuation.domain().hom([~domain.gen()]):
                        raise ValueError("the only allowed automorphism for classical valuations is the automorphism x |--> 1/x")
                    # valuation on the rational function field after x |--> 1/x,
                    # i.e., the classical valuation at infinity
                    return parent.__make_element_class__(InfiniteRationalFunctionFieldValuation)(parent)

                from sage.structure.dynamic_class import dynamic_class
                clazz = RationalFunctionFieldMappedValuation
                if valuation.is_discrete_valuation():
                    clazz = dynamic_class("RationalFunctionFieldMappedValuation_discrete", (clazz, DiscreteValuation))
                else:
                    clazz = dynamic_class("RationalFunctionFieldMappedValuation_infinite", (clazz, InfiniteDiscretePseudoValuation))
                return parent.__make_element_class__(clazz)(parent, valuation, to_valuation_domain, from_valuation_domain)
            return parent.__make_element_class__(FunctionFieldExtensionMappedValuation)(parent, valuation, to_valuation_domain, from_valuation_domain)

        if domain is valuation.domain():
            # we cannot just return valuation in this case
            # as this would break uniqueness and pickling
            raise ValueError("valuation must not be a valuation on domain yet but %r is a valuation on %r" % (valuation, domain))

        if domain.base_field() is domain:
            # valuation is a base valuation on K[x] that induces a valuation on K(x)
            if valuation.restriction(domain.constant_base_field()).is_trivial() and valuation.is_discrete_valuation():
                # valuation corresponds to a finite place
                return parent.__make_element_class__(FiniteRationalFunctionFieldValuation)(parent, valuation)
            from sage.structure.dynamic_class import dynamic_class
            clazz = NonClassicalRationalFunctionFieldValuation
            if valuation.is_discrete_valuation():
                clazz = dynamic_class("NonClassicalRationalFunctionFieldValuation_discrete", (clazz, DiscreteFunctionFieldValuation_base))
            else:
                clazz = dynamic_class("NonClassicalRationalFunctionFieldValuation_negative_infinite", (clazz, NegativeInfiniteDiscretePseudoValuation))
            return parent.__make_element_class__(clazz)(parent, valuation)
        # valuation is a limit valuation that singles out an extension
        return parent.__make_element_class__(FunctionFieldFromLimitValuation)(parent, valuation, domain.polynomial(), extra_args['approximants'])

        raise NotImplementedError("valuation on %r from %r on %r" % (domain, valuation, valuation.domain()))


FunctionFieldValuation = FunctionFieldValuationFactory("sage.rings.function_field.valuation.FunctionFieldValuation")


class FunctionFieldValuation_base(DiscretePseudoValuation):
    r"""
    Abstract base class for any discrete (pseudo-)valuation on a function
    field.

    TESTS::

        sage: K.<x> = FunctionField(QQ)
        sage: v = K.valuation(x)  # indirect doctest
        sage: from sage.rings.function_field.valuation import FunctionFieldValuation_base
        sage: isinstance(v, FunctionFieldValuation_base)
        True
    """


class DiscreteFunctionFieldValuation_base(DiscreteValuation):
    r"""
    Base class for discrete valuations on function fields.

    TESTS::

        sage: K.<x> = FunctionField(QQ)
        sage: v = K.valuation(x)  # indirect doctest
        sage: from sage.rings.function_field.valuation import DiscreteFunctionFieldValuation_base
        sage: isinstance(v, DiscreteFunctionFieldValuation_base)
        True
    """
    def extensions(self, L):
        r"""
        Return the extensions of this valuation to ``L``.

        EXAMPLES::

            sage: K.<x> = FunctionField(QQ)
            sage: v = K.valuation(x)
            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y^2 - x)                                          # needs sage.rings.function_field
            sage: v.extensions(L)                                                       # needs sage.rings.function_field
            [(x)-adic valuation]

        TESTS:

        Valuations over the infinite place::

            sage: v = K.valuation(1/x)
            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y^2 - 1/(x^2 + 1))                                # needs sage.rings.function_field
            sage: sorted(v.extensions(L), key=str)                                      # needs sage.rings.function_field
            [[ Valuation at the infinite place, v(y + 1/x) = 3 ]-adic valuation,
             [ Valuation at the infinite place, v(y - 1/x) = 3 ]-adic valuation]

        Iterated extensions over the infinite place::

            sage: K.<x> = FunctionField(GF(2))
            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y^2 + y + x^3)
            sage: v = K.valuation(1/x)
            sage: w = v.extension(L)
            sage: R.<z> = L[]
            sage: M.<z> = L.extension(z^2 - y)
            sage: w.extension(M)                # not implemented
            Traceback (most recent call last):
            ...
            NotImplementedError

        A case that caused some trouble at some point::

            sage: R.<x> = QQ[]
            sage: v = GaussValuation(R, QQ.valuation(2))

            sage: K.<x> = FunctionField(QQ)
            sage: v = K.valuation(v)

            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y^3 - x^4 - 1)                                    # needs sage.rings.function_field
            sage: v.extensions(L)                                                       # needs sage.rings.function_field
            [2-adic valuation]

        Test that this works in towers::

            sage: K.<x> = FunctionField(GF(2))
            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y - x)
            sage: R.<z> = L[]
            sage: L.<z> = L.extension(z - y)
            sage: v = K.valuation(x)
            sage: v.extensions(L)
            [(x)-adic valuation]
        """
        K = self.domain()
        from sage.categories.function_fields import FunctionFields
        if L is K:
            return [self]
        if L in FunctionFields():
            if K.is_subring(L):
                if L.base() is K:
                    # L = K[y]/(G) is a simple extension of the domain of this valuation
                    G = L.polynomial()
                    if not G.is_monic():
                        G = G / G.leading_coefficient()
                    if any(self(c) < 0 for c in G.coefficients()):
                        # rewrite L = K[u]/(H) with H integral and compute the extensions
                        from sage.rings.valuation.gauss_valuation import GaussValuation
                        g = GaussValuation(G.parent(), self)
                        y_to_u, u_to_y, H = g.monic_integral_model(G)
                        M = K.extension(H, names=L.variable_names())
                        H_extensions = self.extensions(M)

                        from sage.rings.morphism import RingHomomorphism_im_gens
                        if isinstance(y_to_u, RingHomomorphism_im_gens) and isinstance(u_to_y, RingHomomorphism_im_gens):
                            return [L.valuation((w, L.hom([M(y_to_u(y_to_u.domain().gen()))]), M.hom([L(u_to_y(u_to_y.domain().gen()))]))) for w in H_extensions]
                        raise NotImplementedError
                    return [L.valuation(w) for w in self.mac_lane_approximants(L.polynomial(), require_incomparability=True)]
                if L.base() is not L and K.is_subring(L):
                    # recursively call this method for the tower of fields
                    from functools import reduce
                    from operator import add
                    A = [base_valuation.extensions(L) for base_valuation in self.extensions(L.base())]
                    return reduce(add, A, [])
                if L.constant_base_field() is not K.constant_base_field() and K.constant_base_field().is_subring(L):
                    # subclasses should override this method and handle this case, so we never get here
                    raise NotImplementedError("Cannot compute the extensions of %r from %r to %r since the base ring changes." % (self, self.domain(), L))
        raise NotImplementedError("extension of %r from %r to %r not implemented" % (self, K, L))


class RationalFunctionFieldValuation_base(FunctionFieldValuation_base):
    r"""
    Base class for valuations on rational function fields.

    TESTS::

        sage: K.<x> = FunctionField(GF(2))
        sage: v = K.valuation(x)  # indirect doctest
        sage: from sage.rings.function_field.valuation import RationalFunctionFieldValuation_base
        sage: isinstance(v, RationalFunctionFieldValuation_base)
        True
    """
    @cached_method
    def element_with_valuation(self, s):
        r"""
        Return an element with valuation ``s``.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: K.<a> = NumberField(x^3 + 6)
            sage: v = K.valuation(2)
            sage: R.<x> = K[]
            sage: w = GaussValuation(R, v).augmentation(x, 1/123)
            sage: K.<x> = FunctionField(K)
            sage: w = w.extension(K)
            sage: w.element_with_valuation(122/123)
            2/x
            sage: w.element_with_valuation(1)
            2
        """
        constant_valuation = self.restriction(self.domain().constant_base_field())
        if constant_valuation.is_trivial():
            return super().element_with_valuation(s)

        a, b = self.value_group()._element_with_valuation(constant_valuation.value_group(), s)
        ret = self.uniformizer()**a * constant_valuation.element_with_valuation(constant_valuation.value_group().gen() * b)

        return self.simplify(ret, error=s)


class ClassicalFunctionFieldValuation_base(DiscreteFunctionFieldValuation_base):
    r"""
    Base class for discrete valuations on rational function fields that come
    from points on the projective line.

    TESTS::

        sage: K.<x> = FunctionField(GF(5))
        sage: v = K.valuation(x)  # indirect doctest
        sage: from sage.rings.function_field.valuation import ClassicalFunctionFieldValuation_base
        sage: isinstance(v, ClassicalFunctionFieldValuation_base)
        True
    """
    def _test_classical_residue_field(self, **options) -> None:
        r"""
        Check correctness of the residue field of a discrete valuation at a
        classical point.

        TESTS::

            sage: K.<x> = FunctionField(QQ)
            sage: v = K.valuation(x^2 + 1)
            sage: v._test_classical_residue_field()                                     # needs sage.rings.number_field
        """
        tester = self._tester(**options)

        tester.assertTrue(self.domain().constant_base_field().is_subring(self.residue_field()))

    def _ge_(self, other):
        r"""
        Return whether ``self`` is greater or equal to ``other`` everywhere.

        EXAMPLES::

            sage: K.<x> = FunctionField(QQ)
            sage: v = K.valuation(x^2 + 1)
            sage: w = K.valuation(x)
            sage: v >= w
            False
            sage: w >= v
            False
        """
        if other.is_trivial():
            return other.is_discrete_valuation()
        if isinstance(other, ClassicalFunctionFieldValuation_base):
            return self == other
        super()._ge_(other)


class InducedRationalFunctionFieldValuation_base(FunctionFieldValuation_base):
    r"""
    Base class for function field valuation induced by a valuation on the
    underlying polynomial ring.

    TESTS::

        sage: K.<x> = FunctionField(QQ)
        sage: v = K.valuation(x^2 + 1) # indirect doctest
    """
    def __init__(self, parent, base_valuation) -> None:
        r"""
        TESTS::

            sage: K.<x> = FunctionField(QQ)
            sage: v = K.valuation(x) # indirect doctest
            sage: from sage.rings.function_field.valuation import InducedRationalFunctionFieldValuation_base
            sage: isinstance(v, InducedRationalFunctionFieldValuation_base)
            True
        """
        FunctionFieldValuation_base.__init__(self, parent)

        domain = parent.domain()
        if base_valuation.domain() is not domain._ring:
            raise ValueError("base valuation must be defined on %r but %r is defined on %r" % (domain._ring, base_valuation, base_valuation.domain()))

        self._base_valuation = base_valuation

    def uniformizer(self):
        r"""
        Return a uniformizing element for this valuation.

        EXAMPLES::

            sage: K.<x> = FunctionField(QQ)
            sage: K.valuation(x).uniformizer()
            x
        """
        return self.domain()(self._base_valuation.uniformizer())

    def lift(self, F):
        r"""
        Return a lift of ``F`` to the domain of this valuation such
        that :meth:`reduce` returns the original element.

        EXAMPLES::

            sage: K.<x> = FunctionField(QQ)
            sage: v = K.valuation(x)
            sage: v.lift(0)
            0
            sage: v.lift(1)
            1
        """
        F = self.residue_ring().coerce(F)
        if F in self._base_valuation.residue_ring():
            num = self._base_valuation.residue_ring()(F)
            den = self._base_valuation.residue_ring()(1)
        elif F in self._base_valuation.residue_ring().fraction_field():
            num = self._base_valuation.residue_ring()(F.numerator())
            den = self._base_valuation.residue_ring()(F.denominator())
        else:
            raise NotImplementedError("lifting not implemented for this valuation")

        return self.domain()(self._base_valuation.lift(num)) / self.domain()(self._base_valuation.lift(den))

    def value_group(self):
        r"""
        Return the value group of this valuation.

        EXAMPLES::

            sage: K.<x> = FunctionField(QQ)
            sage: K.valuation(x).value_group()
            Additive Abelian Group generated by 1
        """
        return self._base_valuation.value_group()

    def reduce(self, f):
        r"""
        Return the reduction of ``f`` in :meth:`residue_ring`.

        EXAMPLES::

            sage: K.<x> = FunctionField(QQ)
            sage: v = K.valuation(x^2 + 1)
            sage: v.reduce(x)                                                           # needs sage.rings.number_field
            u1
        """
        f = self.domain().coerce(f)

        if self(f) > 0:
            return self.residue_field().zero()
        if self(f) < 0:
            raise ValueError("cannot reduce element of negative valuation")

        base = self._base_valuation

        num = f.numerator()
        den = f.denominator()

        assert base(num) == base(den)
        shift = base.element_with_valuation(-base(num))
        num *= shift
        den *= shift
        ret = base.reduce(num) / base.reduce(den)
        assert not ret.is_zero()
        return self.residue_field()(ret)

    def _repr_(self) -> str:
        r"""
        Return a printable representation of this valuation.

        EXAMPLES::

            sage: K.<x> = FunctionField(QQ)
            sage: K.valuation(x^2 + 1)  # indirect doctest
            (x^2 + 1)-adic valuation
        """
        from sage.rings.valuation.augmented_valuation import AugmentedValuation_base
        from sage.rings.valuation.gauss_valuation import GaussValuation
        if isinstance(self._base_valuation, AugmentedValuation_base):
            if self._base_valuation._base_valuation == GaussValuation(self.domain()._ring, TrivialValuation(self.domain().constant_base_field())):
                if self._base_valuation._mu == 1:
                    return "(%r)-adic valuation" % (self._base_valuation.phi())
        vK = self._base_valuation.restriction(self._base_valuation.domain().base_ring())
        if self._base_valuation == GaussValuation(self.domain()._ring, vK):
            return repr(vK)
        return "Valuation on rational function field induced by %s" % self._base_valuation

    def extensions(self, L):
        r"""
        Return all extensions of this valuation to ``L`` which has a larger
        constant field than the domain of this valuation.

        EXAMPLES::

            sage: K.<x> = FunctionField(QQ)
            sage: v = K.valuation(x^2 + 1)
            sage: L.<x> = FunctionField(GaussianIntegers().fraction_field())
            sage: v.extensions(L)  # indirect doctest
            [(x - I)-adic valuation, (x + I)-adic valuation]
        """
        K = self.domain()
        if L is K:
            return [self]

        from sage.categories.function_fields import FunctionFields
        if (L in FunctionFields()
            and K.is_subring(L)
            and L.base() is L
            and L.constant_base_field() is not K.constant_base_field()
            and K.constant_base_field().is_subring(L.constant_base_field())):
            # The above condition checks whether L is an extension of K that
            # comes from an extension of the field of constants
            # Condition "L.base() is L" is important so we do not call this
            # code for extensions from K(x) to K(x)(y)

            # We extend the underlying valuation on the polynomial ring
            W = self._base_valuation.extensions(L._ring)
            return [L.valuation(w) for w in W]

        return super().extensions(L)

    def _call_(self, f):
        r"""
        Evaluate this valuation at the function ``f``.

        EXAMPLES::

            sage: K.<x> = FunctionField(QQ)
            sage: v = K.valuation(x)  # indirect doctest
            sage: v((x+1)/x^2)
            -2
        """
        return self._base_valuation(f.numerator()) - self._base_valuation(f.denominator())

    def residue_ring(self):
        r"""
        Return the residue field of this valuation.

        EXAMPLES::

            sage: K.<x> = FunctionField(QQ)
            sage: K.valuation(x).residue_ring()
            Rational Field
        """
        return self._base_valuation.residue_ring().fraction_field()

    def restriction(self, ring):
        r"""
        Return the restriction of this valuation to ``ring``.

        EXAMPLES::

            sage: K.<x> = FunctionField(QQ)
            sage: K.valuation(x).restriction(QQ)
            Trivial valuation on Rational Field
        """
        if ring.is_subring(self._base_valuation.domain()):
            return self._base_valuation.restriction(ring)
        return super().restriction(ring)

    def simplify(self, f, error=None, force: bool = False):
        r"""
        Return a simplified version of ``f``.

        Produce an element which differs from ``f`` by an element of
        valuation strictly greater than the valuation of ``f`` (or strictly
        greater than ``error`` if set.)

        If ``force`` is not set, then expensive simplifications may be avoided.

        EXAMPLES::

            sage: K.<x> = FunctionField(QQ)
            sage: v = K.valuation(2)
            sage: f = (x + 1)/(x - 1)

        As the coefficients of this fraction are small, we do not simplify as
        this could be very costly in some cases::

            sage: v.simplify(f)
            (x + 1)/(x - 1)

        However, simplification can be forced::

            sage: v.simplify(f, force=True)
            3
        """
        f = self.domain().coerce(f)

        if error is None:
            # if the caller was sure that we should simplify, then we should try to do the best simplification possible
            error = self(f) if force else self.upper_bound(f)

        from sage.rings.infinity import infinity
        if error is infinity:
            return f

        numerator = f.numerator()
        denominator = f.denominator()

        v_numerator = self._base_valuation(numerator)
        v_denominator = self._base_valuation(denominator)

        if v_numerator - v_denominator > error:
            return self.domain().zero()

        if error == -infinity:
            # This case is not implemented yet, so we just return f which is always safe.
            return f

        numerator = self.domain()(self._base_valuation.simplify(numerator, error=error + v_denominator, force=force))
        denominator = self.domain()(self._base_valuation.simplify(denominator, error=max(v_denominator, error - v_numerator + 2 * v_denominator), force=force))

        ret = numerator / denominator
        assert self(ret - f) > error
        return ret

    def _relative_size(self, f):
        r"""
        Return an estimate on the coefficient size of ``f``.

        The number returned is an estimate on the factor between the number of
        bits used by ``f`` and the minimal number of bits used by an element
        congruent to ``f``.

        This can be used by :meth:`simplify` to decide whether simplification
        of coefficients is going to lead to a significant shrinking of the
        coefficients of ``f``.

        EXAMPLES::

            sage: K.<x> = FunctionField(QQ)
            sage: v = K.valuation(0)
            sage: f = (x + 1024)/(x - 1024)

        Here we report a small size, as the numerator and the denominator
        independently cannot be simplified much::

            sage: v._relative_size(f)
            1

        However, a forced simplification, finds that we could have saved many
        more bits::

            sage: v.simplify(f, force=True)
            -1
        """
        return max(self._base_valuation._relative_size(f.numerator()), self._base_valuation._relative_size(f.denominator()))


class FiniteRationalFunctionFieldValuation(InducedRationalFunctionFieldValuation_base, ClassicalFunctionFieldValuation_base, RationalFunctionFieldValuation_base):
    r"""
    Valuation of a finite place of a function field.

    EXAMPLES::

        sage: K.<x> = FunctionField(QQ)
        sage: v = K.valuation(x + 1); v  # indirect doctest
        (x + 1)-adic valuation

    A finite place with residual degree::

        sage: w = K.valuation(x^2 + 1); w
        (x^2 + 1)-adic valuation

    A finite place with ramification::

        sage: K.<t> = FunctionField(GF(3))
        sage: L.<x> = FunctionField(K)
        sage: u = L.valuation(x^3 - t); u
        (x^3 + 2*t)-adic valuation

    A finite place with residual degree and ramification::

        sage: q = L.valuation(x^6 - t); q
        (x^6 + 2*t)-adic valuation
    """
    def __init__(self, parent, base_valuation) -> None:
        r"""
        TESTS::

            sage: K.<x> = FunctionField(QQ)
            sage: v = K.valuation(x + 1)
            sage: from sage.rings.function_field.valuation import FiniteRationalFunctionFieldValuation
            sage: isinstance(v, FiniteRationalFunctionFieldValuation)
            True
        """
        InducedRationalFunctionFieldValuation_base.__init__(self, parent, base_valuation)
        ClassicalFunctionFieldValuation_base.__init__(self, parent)
        RationalFunctionFieldValuation_base.__init__(self, parent)


class NonClassicalRationalFunctionFieldValuation(InducedRationalFunctionFieldValuation_base, RationalFunctionFieldValuation_base):
    r"""
    Valuation induced by a valuation on the underlying polynomial ring which is
    non-classical.

    EXAMPLES::

        sage: K.<x> = FunctionField(QQ)
        sage: v = GaussValuation(QQ['x'], QQ.valuation(2))
        sage: w = K.valuation(v); w  # indirect doctest
        2-adic valuation
    """
    def __init__(self, parent, base_valuation) -> None:
        r"""
        TESTS:

        There is some support for discrete pseudo-valuations on rational
        function fields in the code. However, since these valuations must send
        elements to `-\infty`, they are not supported yet::

            sage: R.<x> = QQ[]
            sage: v = GaussValuation(QQ['x'], QQ.valuation(2)).augmentation(x, infinity)
            sage: K.<x> = FunctionField(QQ)
            sage: w = K.valuation(v)
            sage: from sage.rings.function_field.valuation import NonClassicalRationalFunctionFieldValuation
            sage: isinstance(w, NonClassicalRationalFunctionFieldValuation)
            True
        """
        InducedRationalFunctionFieldValuation_base.__init__(self, parent, base_valuation)
        RationalFunctionFieldValuation_base.__init__(self, parent)

    def residue_ring(self):
        r"""
        Return the residue field of this valuation.

        EXAMPLES::

            sage: K.<x> = FunctionField(QQ)
            sage: v = valuations.GaussValuation(QQ['x'], QQ.valuation(2))
            sage: w = K.valuation(v)
            sage: w.residue_ring()
            Rational function field in x over Finite Field of size 2

            sage: R.<x> = QQ[]
            sage: vv = v.augmentation(x, 1)
            sage: w = K.valuation(vv)
            sage: w.residue_ring()
            Rational function field in x over Finite Field of size 2

            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y^2 + 2*x)                                        # needs sage.rings.function_field
            sage: w.extension(L).residue_ring()                                         # needs sage.rings.function_field
            Function field in u2 defined by u2^2 + x

        TESTS:

        This still works for pseudo-valuations::

            sage: R.<x> = QQ[]
            sage: v = valuations.GaussValuation(R, QQ.valuation(2))
            sage: vv = v.augmentation(x, infinity)
            sage: K.<x> = FunctionField(QQ)
            sage: w = K.valuation(vv)
            sage: w.residue_ring()
            Finite Field of size 2
        """
        if not self.is_discrete_valuation():
            # A pseudo valuation attaining negative infinity does typically not have a function field as its residue ring
            return super().residue_ring()
        return self._base_valuation.residue_ring().fraction_field().function_field()


class FunctionFieldFromLimitValuation(FiniteExtensionFromLimitValuation, DiscreteFunctionFieldValuation_base):
    r"""
    A valuation on a finite extensions of function fields `L=K[y]/(G)` where `K` is
    another function field.

    EXAMPLES::

        sage: K.<x> = FunctionField(QQ)
        sage: R.<y> = K[]
        sage: L.<y> = K.extension(y^2 - (x^2 + x + 1))                                  # needs sage.rings.function_field
        sage: v = K.valuation(x - 1)  # indirect doctest                                # needs sage.rings.function_field
        sage: w = v.extension(L); w                                                     # needs sage.rings.function_field
        (x - 1)-adic valuation
    """
    def __init__(self, parent, approximant, G, approximants) -> None:
        r"""
        TESTS::

            sage: K.<x> = FunctionField(QQ)
            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y^2 - (x^2 + x + 1))
            sage: v = K.valuation(x - 1) # indirect doctest
            sage: w = v.extension(L)
            sage: from sage.rings.function_field.valuation import FunctionFieldFromLimitValuation
            sage: isinstance(w, FunctionFieldFromLimitValuation)
            True
        """
        FiniteExtensionFromLimitValuation.__init__(self, parent, approximant, G, approximants)
        DiscreteFunctionFieldValuation_base.__init__(self, parent)

    def _to_base_domain(self, f):
        r"""
        Return ``f`` as an element of the domain of the underlying limit valuation.

        EXAMPLES::

            sage: K.<x> = FunctionField(QQ)
            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y^2 - (x^2 + x + 1))
            sage: v = K.valuation(x - 1) # indirect doctest
            sage: w = v.extension(L)
            sage: w._to_base_domain(y).parent()
            Univariate Polynomial Ring in y over Rational function field in x over Rational Field
        """
        return f.element()

    def scale(self, scalar):
        r"""
        Return this valuation scaled by ``scalar``.

        EXAMPLES::

            sage: K.<x> = FunctionField(QQ)
            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y^2 - (x^2 + x + 1))
            sage: v = K.valuation(x - 1) # indirect doctest
            sage: w = v.extension(L)
            sage: 3*w
            3 * (x - 1)-adic valuation
        """
        if scalar in QQ and scalar > 0 and scalar != 1:
            return self.domain().valuation(self._base_valuation._initial_approximation.scale(scalar))
        return super().scale(scalar)


class FunctionFieldMappedValuation_base(FunctionFieldValuation_base, MappedValuation_base):
    r"""
    A valuation on a function field which relies on a ``base_valuation`` on an
    isomorphic function field.

    EXAMPLES::

        sage: K.<x> = FunctionField(GF(2))
        sage: v = K.valuation(1/x); v
        Valuation at the infinite place
    """
    def __init__(self, parent, base_valuation, to_base_valuation_domain, from_base_valuation_domain) -> None:
        r"""
        TESTS::

            sage: K.<x> = FunctionField(GF(2))
            sage: v = K.valuation(1/x)
            sage: from sage.rings.function_field.valuation import FunctionFieldMappedValuation_base
            sage: isinstance(v, FunctionFieldMappedValuation_base)
            True
        """
        FunctionFieldValuation_base.__init__(self, parent)
        MappedValuation_base.__init__(self, parent, base_valuation)

        self._to_base = to_base_valuation_domain
        self._from_base = from_base_valuation_domain

    def _to_base_domain(self, f):
        r"""
        Return ``f`` as an element in the domain of ``_base_valuation``.

        EXAMPLES::

            sage: K.<x> = FunctionField(GF(2))
            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y^2 + y + x^3)                                    # needs sage.rings.function_field
            sage: v = K.valuation(1/x)
            sage: w = v.extension(L)                                                    # needs sage.rings.function_field
            sage: w._to_base_domain(y)                                                  # needs sage.rings.function_field
            x^2*y
        """
        return self._to_base(f)

    def _from_base_domain(self, f):
        r"""
        Return ``f`` as an element in the domain of this valuation.

        EXAMPLES::

            sage: K.<x> = FunctionField(GF(2))
            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y^2 + y + x^3)                                    # needs sage.rings.function_field
            sage: v = K.valuation(1/x)
            sage: w = v.extension(L)                                                    # needs sage.rings.function_field
            sage: w._from_base_domain(w._to_base_domain(y))                             # needs sage.rings.function_field
            y

        r"""
        return self._from_base(f)

    def scale(self, scalar):
        r"""
        Return this valuation scaled by ``scalar``.

        EXAMPLES::

            sage: K.<x> = FunctionField(GF(2))
            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y^2 + y + x^3)                                    # needs sage.rings.function_field
            sage: v = K.valuation(1/x)
            sage: w = v.extension(L)                                                    # needs sage.rings.function_field
            sage: 3*w                                                                   # needs sage.rings.function_field
            3 * (x)-adic valuation (in Rational function field in x over Finite Field of size 2 after x |--> 1/x)
        """
        from sage.rings.rational_field import QQ
        if scalar in QQ and scalar > 0 and scalar != 1:
            return self.domain().valuation((self._base_valuation.scale(scalar), self._to_base, self._from_base))
        return super().scale(scalar)

    def _repr_(self) -> str:
        r"""
        Return a printable representation of this valuation.

        EXAMPLES::

            sage: K.<x> = FunctionField(GF(2))
            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y^2 + y + x^3)                                    # needs sage.rings.function_field
            sage: v = K.valuation(1/x)
            sage: v.extension(L)  # indirect doctest                                    # needs sage.rings.function_field
            Valuation at the infinite place
        """
        to_base = repr(self._to_base)
        if hasattr(self._to_base, '_repr_defn'):
            to_base = self._to_base._repr_defn().replace('\n', ', ')
        return "%r (in %r after %s)" % (self._base_valuation, self._base_valuation.domain(), to_base)

    def is_discrete_valuation(self):
        r"""
        Return whether this is a discrete valuation.

        EXAMPLES::

            sage: K.<x> = FunctionField(QQ)
            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y^2 - x^4 - 1)
            sage: v = K.valuation(1/x)
            sage: w0,w1 = v.extensions(L)
            sage: w0.is_discrete_valuation()
            True
        """
        return self._base_valuation.is_discrete_valuation()


class FunctionFieldMappedValuationRelative_base(FunctionFieldMappedValuation_base):
    r"""
    A valuation on a function field which relies on a ``base_valuation`` on an
    isomorphic function field and which is such that the map from and to the
    other function field is the identity on the constant field.

    EXAMPLES::

        sage: K.<x> = FunctionField(GF(2))
        sage: v = K.valuation(1/x); v
        Valuation at the infinite place
    """
    def __init__(self, parent, base_valuation, to_base_valuation_domain, from_base_valuation_domain) -> None:
        r"""
        TESTS::

            sage: K.<x> = FunctionField(GF(2))
            sage: v = K.valuation(1/x)
            sage: from sage.rings.function_field.valuation import FunctionFieldMappedValuationRelative_base
            sage: isinstance(v, FunctionFieldMappedValuationRelative_base)
            True
        """
        FunctionFieldMappedValuation_base.__init__(self, parent, base_valuation, to_base_valuation_domain, from_base_valuation_domain)
        if self.domain().constant_base_field() is not base_valuation.domain().constant_base_field():
            raise ValueError("constant fields must be identical but they differ for %r and %r" % (self.domain(), base_valuation.domain()))

    def restriction(self, ring):
        r"""
        Return the restriction of this valuation to ``ring``.

        EXAMPLES::

            sage: K.<x> = FunctionField(GF(2))
            sage: K.valuation(1/x).restriction(GF(2))
            Trivial valuation on Finite Field of size 2
        """
        if ring.is_subring(self.domain().constant_base_field()):
            return self._base_valuation.restriction(ring)
        return super().restriction(ring)


class RationalFunctionFieldMappedValuation(FunctionFieldMappedValuationRelative_base, RationalFunctionFieldValuation_base):
    r"""
    Valuation on a rational function field that is implemented after a map to
    an isomorphic rational function field.

    EXAMPLES::

        sage: K.<x> = FunctionField(QQ)
        sage: R.<x> = QQ[]
        sage: w = GaussValuation(R, QQ.valuation(2)).augmentation(x, 1)
        sage: w = K.valuation(w)
        sage: v = K.valuation((w, K.hom([~K.gen()]), K.hom([~K.gen()]))); v
        Valuation on rational function field induced by
        [ Gauss valuation induced by 2-adic valuation, v(x) = 1 ]
        (in Rational function field in x over Rational Field after x |--> 1/x)
    """
    def __init__(self, parent, base_valuation, to_base_valuation_doain, from_base_valuation_domain) -> None:
        r"""
        TESTS::

            sage: K.<x> = FunctionField(QQ)
            sage: R.<x> = QQ[]
            sage: w = GaussValuation(R, QQ.valuation(2)).augmentation(x, 1)
            sage: w = K.valuation(w)
            sage: v = K.valuation((w, K.hom([~K.gen()]), K.hom([~K.gen()])))
            sage: from sage.rings.function_field.valuation import RationalFunctionFieldMappedValuation
            sage: isinstance(v, RationalFunctionFieldMappedValuation)
            True
        """
        FunctionFieldMappedValuationRelative_base.__init__(self, parent, base_valuation, to_base_valuation_doain, from_base_valuation_domain)
        RationalFunctionFieldValuation_base.__init__(self, parent)


class InfiniteRationalFunctionFieldValuation(FunctionFieldMappedValuationRelative_base, RationalFunctionFieldValuation_base, ClassicalFunctionFieldValuation_base):
    r"""
    Valuation of the infinite place of a function field.

    EXAMPLES::

        sage: K.<x> = FunctionField(QQ)
        sage: v = K.valuation(1/x)  # indirect doctest
    """
    def __init__(self, parent) -> None:
        r"""
        TESTS::

            sage: K.<x> = FunctionField(QQ)
            sage: v = K.valuation(1/x)  # indirect doctest
            sage: from sage.rings.function_field.valuation import InfiniteRationalFunctionFieldValuation
            sage: isinstance(v, InfiniteRationalFunctionFieldValuation)
            True
        """
        x = parent.domain().gen()
        FunctionFieldMappedValuationRelative_base.__init__(self, parent, FunctionFieldValuation(parent.domain(), x), parent.domain().hom([1 / x]), parent.domain().hom([1 / x]))
        RationalFunctionFieldValuation_base.__init__(self, parent)
        ClassicalFunctionFieldValuation_base.__init__(self, parent)

    def _repr_(self) -> str:
        r"""
        Return a printable representation of this valuation.

        EXAMPLES::

            sage: K.<x> = FunctionField(QQ)
            sage: K.valuation(1/x)  # indirect doctest
            Valuation at the infinite place
        """
        return "Valuation at the infinite place"


class FunctionFieldExtensionMappedValuation(FunctionFieldMappedValuationRelative_base):
    r"""
    A valuation on a finite extensions of function fields `L=K[y]/(G)` where `K` is
    another function field which redirects to another ``base_valuation`` on an
    isomorphism function field `M=K[y]/(H)`.

    The isomorphisms must be trivial on ``K``.

    EXAMPLES::

        sage: K.<x> = FunctionField(GF(2))
        sage: R.<y> = K[]
        sage: L.<y> = K.extension(y^2 + y + x^3)                                        # needs sage.rings.function_field
        sage: v = K.valuation(1/x)
        sage: w = v.extension(L)                                                        # needs sage.rings.function_field

        sage: w(x)                                                                      # needs sage.rings.function_field
        -1
        sage: w(y)                                                                      # needs sage.rings.function_field
        -3/2
        sage: w.uniformizer()                                                           # needs sage.rings.function_field
        1/x^2*y

    TESTS::

        sage: from sage.rings.function_field.valuation import FunctionFieldExtensionMappedValuation
        sage: isinstance(w, FunctionFieldExtensionMappedValuation)                      # needs sage.rings.function_field
        True
    """
    def _repr_(self) -> str:
        r"""
        Return a printable representation of this valuation.

        EXAMPLES::

            sage: K.<x> = FunctionField(GF(2))
            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y^2 + y + x^3)
            sage: v = K.valuation(1/x)
            sage: w = v.extension(L); w
            Valuation at the infinite place

            sage: K.<x> = FunctionField(QQ)
            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y^2 - 1/x^2 - 1)
            sage: v = K.valuation(1/x)
            sage: w = v.extensions(L); w
            [[ Valuation at the infinite place, v(y + 1) = 2 ]-adic valuation,
             [ Valuation at the infinite place, v(y - 1) = 2 ]-adic valuation]
        """
        assert (self.domain().base() is not self.domain())
        if repr(self._base_valuation) == repr(self.restriction(self.domain().base())):
            return repr(self._base_valuation)
        return super()._repr_()

    def restriction(self, ring):
        r"""
        Return the restriction of this valuation to ``ring``.

        EXAMPLES::

            sage: K.<x> = FunctionField(GF(2))
            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y^2 + y + x^3)
            sage: v = K.valuation(1/x)
            sage: w = v.extension(L)
            sage: w.restriction(K) is v
            True
        """
        if ring.is_subring(self.domain().base()):
            return self._base_valuation.restriction(ring)
        return super().restriction(ring)



"""
Octonion Algebras

AUTHORS:

- Travis Scrimshaw (2023-05-06): Initial version
"""
# ****************************************************************************
#       Copyright (C) 2023 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.misc.functional import sqrt
from sage.structure.element cimport Element, parent
from sage.structure.parent cimport Parent
from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.richcmp cimport richcmp
from sage.structure.category_object import normalize_names
from sage.modules.free_module import FreeModule
from sage.categories.magmatic_algebras import MagmaticAlgebras
from sage.categories.rings import Rings
from sage.categories.metric_spaces import MetricSpaces


cdef class Octonion_generic(AlgebraElement):
    r"""
    An octonion with generic parameters.
    """
    def __init__(self, parent, v):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ, 1, 3, 7)
            sage: elt = sum(i * b for i, b in enumerate(O.basis(), start=2))
            sage: TestSuite(elt).run()
            sage: elt = sum(O.basis())
            sage: TestSuite(elt).run()
            sage: TestSuite(O.zero()).run()

            sage: O = OctonionAlgebra(Zmod(8), 1, 3, 7)
            sage: elt = sum(i * b for i, b in enumerate(O.basis(), start=2))
            sage: TestSuite(elt).run()
            sage: elt = sum(O.basis())
            sage: TestSuite(elt).run()
            sage: TestSuite(O.zero()).run()
        """
        v.set_immutable()
        self.vec = v
        super().__init__(parent)

    def _repr_(self) -> str:
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ, 1, 3, 7)
            sage: elt = sum(i * b for i, b in enumerate(O.basis(), start=2))
            sage: elt
            2 + 3*i + 4*j + 5*k + 6*l + 7*li + 8*lj + 9*lk
        """
        from sage.algebras.weyl_algebra import repr_from_monomials
        data = [p for p in enumerate(self.vec) if p[1]]
        return repr_from_monomials(data, self._parent._repr_term)

    def _latex_(self):
        r"""
        Return a `\LaTeX` representation of ``self``.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ, 1, 3, 7)
            sage: elt = sum(i * b for i, b in enumerate(O.basis(), start=2))
            sage: latex(elt)
            2 + 3 i + 4 j + 5 k + 6 l + 7 li + 8 lj + 9 lk
        """
        from sage.algebras.weyl_algebra import repr_from_monomials
        data = [p for p in enumerate(self.vec) if p[1]]
        return repr_from_monomials(data, self._parent._repr_term, True)

    def __getitem__(self, m):
        r"""
        Return the ``m``-th coordinate of ``self``.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ)
            sage: elt = sum(i * b for i, b in enumerate(O.basis()))
            sage: elt
            i + 2*j + 3*k + 4*l + 5*li + 6*lj + 7*lk
            sage: elt[2]
            2
            sage: elt.coefficient(4)
            4
            sage: elt[6]
            6
            sage: elt.leading_coefficient()
            7
            sage: elt.dense_coefficient_list()
            [0, 1, 2, 3, 4, 5, 6, 7]
        """
        return self.vec[m]

    def __bool__(self):
        r"""
        Return if ``self`` is nonzero or not.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ, 1, 3, 7)
            sage: elt = sum(i * b for i, b in enumerate(O.basis(), start=2))
            sage: bool(elt)
            True
            sage: bool(O.zero())
            False
        """
        return bool(self.vec)

    def __reduce__(self):
        r"""
        For pickling.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ, 1, 3, 7)
            sage: x = sum(i * b for i, b in enumerate(O.basis(), start=2))
            sage: x.__reduce__()
            (<class 'sage.algebras.octonion_algebra.Octonion_generic'>,
             (Octonion algebra over Rational Field with parameters (1, 3, 7),
              (2, 3, 4, 5, 6, 7, 8, 9)))
        """
        return (self.__class__, (self._parent, self.vec))

    cpdef _richcmp_(self, other, int op):
        r"""
        Compare ``self`` to ``other`` with type ``op``.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ, 1, 3, 7)
            sage: x = sum(i * b for i, b in enumerate(O.basis(), start=2))
            sage: y = sum(O.basis())
            sage: x != y
            True
            sage: x == y
            False
            sage: x < y
            False
        """
        return richcmp(self.vec, (<Octonion_generic> other).vec, op)

    def __hash__(self):
        r"""
        Return a hash of ``self``.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ, 1, 3, 7)
            sage: elt = sum(i * b for i, b in enumerate(O.basis(), start=2))
            sage: hash(elt) == hash(elt.vector())
            True
        """
        return hash(self.vec)

    cpdef _add_(self, other):
        r"""
        Return ``self`` plus ``other``.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ, 1, 3, 7)
            sage: x = sum(i * b for i, b in enumerate(O.basis(), start=2))
            sage: y = sum(O.basis())
            sage: x + y
            3 + 4*i + 5*j + 6*k + 7*l + 8*li + 9*lj + 10*lk
        """
        return self.__class__(self._parent, self.vec + (<Octonion_generic> other).vec)

    cpdef _sub_(self, other):
        r"""
        Return ``self`` minus ``other``.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ, 1, 3, 7)
            sage: x = sum(i * b for i, b in enumerate(O.basis(), start=2))
            sage: y = sum(O.basis())
            sage: x - y
            1 + 2*i + 3*j + 4*k + 5*l + 6*li + 7*lj + 8*lk
        """
        return self.__class__(self._parent, self.vec - (<Octonion_generic> other).vec)

    def __neg__(self):
        r"""
        Return the negative of ``self``.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ, 1, 3, 7)
            sage: x = sum(i * b for i, b in enumerate(O.basis(), start=2))
            sage: -x
            -2 - 3*i - 4*j - 5*k - 6*l - 7*li - 8*lj - 9*lk
            sage: y = sum(O.basis())
            sage: -y
            -1 - i - j - k - l - li - lj - lk
        """
        return self.__class__(self._parent, -self.vec)

    cpdef _lmul_(self, Element other):
        r"""
        Return ``self * other`` for a scalar ``other``.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ)
            sage: i, j, k, l = O.gens()
            sage: elt = 2 * i + 3 * j + k * 4 + l * 5
            sage: elt * 5
            10*i + 15*j + 20*k + 25*l
        """
        return self.__class__(self._parent, self.vec * other)

    cpdef _rmul_(self, Element other):
        r"""
        Return ``self * other`` for a scalar ``other``.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ)
            sage: i, j, k, l = O.gens()
            sage: elt = 2 * i + 3 * j + k * 4 + l * 5
            sage: 5 * elt
            10*i + 15*j + 20*k + 25*l
        """
        return self.__class__(self._parent, other * self.vec)

    cpdef _mul_(self, other):
        r"""
        Return ``self`` multiplied by ``other``.

        INPUT:

        - ``other`` -- element of the octonion algebra as ``self``

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ)
            sage: table([[r * c for c in O.basis()] for r in O.basis()])
              1    i     j     k     l     li    lj    lk
              i    -1    -k    j     -li   l     lk    -lj
              j    k     -1    -i    -lj   -lk   l     li
              k    -j    i     -1    -lk   lj    -li   l
              l    li    lj    lk    -1    -i    -j    -k
              li   -l    lk    -lj   i     -1    k     -j
              lj   -lk   -l    li    j     -k    -1    i
              lk   lj    -li   -l    k     j     -i    -1

            sage: SO = OctonionAlgebra(QQ, c=1)
            sage: table([[r * c for c in SO.basis()] for r in SO.basis()])
              1    i     j     k     l     li    lj    lk
              i    -1    -k    j     -li   l     lk    -lj
              j    k     -1    -i    -lj   -lk   l     li
              k    -j    i     -1    -lk   lj    -li   l
              l    li    lj    lk    1     i     j     k
              li   -l    lk    -lj   -i    1     -k    j
              lj   -lk   -l    li    -j    k     1     -i
              lk   lj    -li   -l    -k    -j    i     1
        """
        cdef Parent P = self._parent
        cdef list ret = [P._base.zero()] * 8
        cdef tuple table = <tuple> P._mult_table
        cdef tuple row
        cdef int i, j, k
        cdef FreeModuleElement rv = (<Octonion_generic> other).vec
        for i in range(8):
            row = <tuple> table[i]
            cl = self.vec.get_unsafe(i)
            for j in range(8):
                k = <int> row[j][0]
                coeff = row[j][1]
                cr = rv.get_unsafe(j)
                ret[k] += cl * cr * coeff
        return self.__class__(P, P._module(ret))

    cpdef _div_(self, other):
        """
        Return ``self`` divided by ``other``.

        EXAMPLES::

            sage: O = OctonionAlgebra(ZZ)
            sage: x = O([0, 0, 2, 0, 2, 0, 0, 0])
            sage: y = O([1, 0, 0, 0, 1, 0, 0, 0])
            sage: y.quadratic_form()
            2
            sage: x * y.conjugate()
            2 + 2*j + 2*l + 2*lj
            sage: x / y
            1 + j + l + lj
        """
        cdef Octonion_generic temp = <Octonion_generic> self * (<Octonion_generic> other).conjugate()
        qf = (<Octonion_generic> other).quadratic_form()
        return self.__class__(self._parent, temp.vec / qf)

    def is_unit(self) -> bool:
        r"""
        Return if ``self`` is a unit or not.

        EXAMPLES::

            sage: O = OctonionAlgebra(ZZ)
            sage: x = O([1, 0, 1, 0, 0, 0, 0, 0])
            sage: x.quadratic_form()
            2
            sage: x.is_unit()
            False
            sage: O([1, 0, -1, 0, 0, 0, 0, 0]).is_unit()
            False
            sage: x = O([1, 0, 0, 0, 1, 0, 0, 0])
            sage: x.quadratic_form()
            2
            sage: x.is_unit()
            False
            sage: x = O([0, 0, 0, 0, 1, 0, 0, 0])
            sage: x.quadratic_form()
            1
            sage: x.is_unit()
            True

            sage: O = OctonionAlgebra(ZZ, -1, 1, 2)
            sage: x = O([1, 0, 1, 0, 0, 0, 0, 0])
            sage: x.quadratic_form()
            0
            sage: x.is_unit()
            False
            sage: O([1, 0, -1, 0, 0, 0, 0, 0]).is_unit()
            False
            sage: x = O([1, 0, 0, 0, 1, 0, 0, 0])
            sage: x.quadratic_form()
            -1
            sage: x.is_unit()
            True
            sage: x = O([0, 0, 0, 0, 1, 0, 0, 0])
            sage: x.quadratic_form()
            -2
            sage: x.is_unit()
            False
        """
        return self.quadratic_form().is_unit()

    def __invert__(self):
        r"""
        Return the (multiplicative) inverse of ``self``.

        EXAMPLES::

            sage: O137 = OctonionAlgebra(QQ, 1, 3, 7)
            sage: elt = sum(O137.basis())
            sage: elt.quadratic_form()
            0
            sage: ~elt
            Traceback (most recent call last):
            ...
            ZeroDivisionError: rational division by zero
            sage: ~O137.zero()
            Traceback (most recent call last):
            ...
            ZeroDivisionError
        """
        if not self.vec:
            raise ZeroDivisionError
        return self.quadratic_form().inverse_of_unit() * self.conjugate()

    cpdef Octonion_generic conjugate(self):
        r"""
        Return the conjugate of ``self``.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ)
            sage: elt = sum(O.basis()); elt
            1 + i + j + k + l + li + lj + lk
            sage: elt.conjugate()
            1 - i - j - k - l - li - lj - lk
        """
        cdef FreeModuleElement v = <FreeModuleElement> -self.vec
        v.set_unsafe(0, -v.get_unsafe(0))
        return self.__class__(self._parent, v)

    cpdef quadratic_form(self):
        r"""
        Return the quadratic form of ``self``.

        The octonion algebra has a distinguished quadratic form given
        by `N(x) = x x^*`, where `x^*` is the :meth:`conjugate` of `x`.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ, 1, 3, 7)
            sage: elt = sum(O.basis())
            sage: elt.quadratic_form()
            0
            sage: elt * elt.conjugate()
            0
        """
        cdef int i
        cdef tuple table = self._parent._mult_table
        ret = self.vec.get_unsafe(0) ** 2
        for i in range(1, 8):
            ret += -(<tuple> table[i])[i][1] * self.vec.get_unsafe(i) ** 2
        return ret

    cpdef norm(self):
        r"""
        Return the norm of ``self``.

        The norm of an octonion `x` is `\lVert x \rVert = \sqrt{x x^*}`,
        where `x^*` is the :meth:`conjugate` of `x`.

        .. SEEALSO::

            This is the square root of :meth:`quadratic_form()`.

        .. WARNING::

            If any of the parameters `a, b, c \not> 0`, then this is not
            an actual norm.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ, 1, 3, 7)
            sage: elt = sum(i * b for i, b in enumerate(O.basis(), start=2))
            sage: elt.norm()
            2*sqrt(-61)
            sage: elt = sum(O.basis())
            sage: elt.norm()
            0
        """
        return sqrt(self.quadratic_form())

    cpdef abs(self):
        r"""
        Return the absolute value of ``self``.

        This is equal to the :meth:`norm`.

        .. WARNING::

            If any of the parameters `a, b, c \not> 0`, then this does
            not define a metric.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ, 1, 3, 7)
            sage: elt = sum(i * b for i, b in enumerate(O.basis(), start=2))
            sage: elt.abs()
            2*sqrt(-61)
            sage: elt = sum(O.basis())
            sage: elt.abs()
            0
        """
        return self.norm()

    cpdef real_part(self):
        r"""
        Return the real part of ``self``.

        OUTPUT: the real part of ``self`` as an element in the base ring

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ)
            sage: elt = sum(i * b for i, b in enumerate(O.basis(), start=2)); elt
            2 + 3*i + 4*j + 5*k + 6*l + 7*li + 8*lj + 9*lk
            sage: r = elt.real_part(); r
            2
            sage: r.parent() is QQ
            True
        """
        return self.vec.get_unsafe(0)

    cpdef Octonion_generic imag_part(self):
        r"""
        Return the imginary part of ``self``.

        OUTPUT: the imaginary part of ``self`` as an element in the octonion algebra

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ)
            sage: elt = sum(i * b for i, b in enumerate(O.basis(), start=2)); elt
            2 + 3*i + 4*j + 5*k + 6*l + 7*li + 8*lj + 9*lk
            sage: elt.imag_part()
            3*i + 4*j + 5*k + 6*l + 7*li + 8*lj + 9*lk

        TESTS::

            sage: O = OctonionAlgebra(QQ)
            sage: elt = sum(i * b for i, b in enumerate(O.basis(), start=2))
            sage: elt.imag_part()
            3*i + 4*j + 5*k + 6*l + 7*li + 8*lj + 9*lk
            sage: elt
            2 + 3*i + 4*j + 5*k + 6*l + 7*li + 8*lj + 9*lk
        """
        cdef FreeModuleElement v = <FreeModuleElement> self.vec.__copy__()
        v.set_unsafe(0, self._parent._base.zero())
        return self.__class__(self._parent, v)

    def _vector_(self, new_base_ring=None):
        r"""
        Return ``self`` as a vector in `R^8`, where `R` is the base
        ring of ``self``.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ)
            sage: elt = sum(i * b for i, b in enumerate(O.basis(), start=2))
            sage: elt.vector()
            (2, 3, 4, 5, 6, 7, 8, 9)
        """
        if new_base_ring is None:
            return self.vec
        return self.vec.change_ring(new_base_ring)

    vector = _vector_

    def monomial_coefficients(self, copy=False):
        """
        Return ``self`` as a ``dict`` with keys being indices
        for the basis and the values being the corresponding
        nonzero coefficients.

        INPUT:

        - ``copy`` -- ignored

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ)
            sage: x = O([2/7, 0, 0, 0, 2/3, 0, -5, 0])
            sage: x.monomial_coefficients()
            {0: 2/7, 4: 2/3, 6: -5}
        """
        return self.vec.dict()

    dict = monomial_coefficients


cdef class Octonion(Octonion_generic):
    r"""
    An octonion.

    This is an element of the octonion algebra with parameters
    `a = b = c = -1`, which is a classical octonion number.
    """
    cpdef quadratic_form(self):
        r"""
        Return the quadratic form of ``self``.

        The octonion algebra has a distinguished quadratic form given
        by `N(x) = x x^*`, where `x^*` is the :meth:`conjugate` of `x`.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ)
            sage: elt = sum(O.basis()); elt
            1 + i + j + k + l + li + lj + lk
            sage: elt.quadratic_form()
            8
            sage: elt * elt.conjugate()
            8
        """
        return self.vec * self.vec

    cpdef norm(self):
        r"""
        Return the norm of ``self``.

        The norm of an octonion `x` is `\lVert x \rVert = \sqrt{x x^*}`,
        where `x^*` is the :meth:`conjugate` of `x`.

        .. SEEALSO::

            This is the square root of :meth:`quadratic_form()`.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ)
            sage: elt = sum(i * b for i, b in enumerate(O.basis(), start=2))
            sage: elt.norm()
            2*sqrt(71)
            sage: elt = sum(O.basis())
            sage: elt.norm()
            2*sqrt(2)
        """
        return self.vec.norm()


class OctonionAlgebra(UniqueRepresentation, Parent):
    r"""
    The octonion algebra.

    Let `R` be a commutative ring of characteristic not equal to `2`. The
    *octonion algebra* with parameters `a, b, c` is a non-associative
    non-commutative unital 8-dimensional `R`-algebra that is a deformation
    of the usual octonions, which are when `a = b = c = -1`. The octonions
    were originally constructed by Graves and independently discovered by
    Cayley (due to being published first, these are sometimes called
    the Cayley numbers) and can also be built from the Cayley-Dickson
    construction with the :class:`quaternions <QuaternionAlgebra>`.

    We use the multiplication table from [Scha1996]_. The octonion
    algebra `\mathbf{O}_{a,b,c}(R)` is a composition (Hurwitz) algebra,
    which means it is also an alternative algebra as it satisfies
    `x^2 y = (x x) y = x (x y)` and  `y x^2 = y (x x) = (y x) x`
    for all `x, y \in \mathbf{O}_{a,b,c}`.

    EXAMPLES:

    We first create the classical octonions and perform some basic
    computations::

        sage: O = OctonionAlgebra(QQ)
        sage: O
        Octonion algebra over Rational Field
        sage: i, j, k, l = O.gens()
        sage: i * j * k
        1
        sage: k * j * i
        -1
        sage: (i * k) * l
        -lj
        sage: i * (k * l)
        lj
        sage: elt = sum(O.basis())
        sage: elt^2
        -6 + 2*i + 2*j + 2*k + 2*l + 2*li + 2*lj + 2*lk
        sage: prod(O.basis())
        1
        sage: (i + l)^2
        -2
        sage: (1 + l) * (1 + l).conjugate()
        2
        sage: S = O.some_elements()
        sage: B = O.basis()
        sage: S.extend(x * (i + j/2 - 5*k/3) for x in O.some_elements())
        sage: all((x * x) * y == (x * (x * y)) for x in S for y in S)
        True
        sage: all(y * (x * x) == (y * x) * x for x in S for y in S)
        True
        sage: all((x + x.conjugate()) / 2 == x.real_part() for x in S)
        True
        sage: all((x - x.conjugate()) / 2 == x.imag_part() for x in S)
        True
        sage: all(sum((b*x)*b for b in B) == -6 * x.conjugate() for x in S)
        True

    We construct the (rescaled) `E_8` lattice as the integral octonions,
    which we verify by constructing `240` shortest length elements in
    the lattice (see also :wikipedia:`E8_lattice#Integral_octonions`)::

        sage: m = (i + j + k + l) / 2
        sage: basis = [i, j, i*j, i^2, m, i * m, j * m, (i * j) * m]
        sage: basis
        [i,
         j,
         -k,
         -1,
         1/2*i + 1/2*j + 1/2*k + 1/2*l,
         -1/2 + 1/2*j - 1/2*k - 1/2*li,
         -1/2 - 1/2*i + 1/2*k - 1/2*lj,
         1/2 - 1/2*i + 1/2*j + 1/2*lk]
        sage: matrix([vector(b) for b in basis]).rank()
        8
        sage: [b.norm() for b in basis]
        [1, 1, 1, 1, 1, 1, 1, 1]
        sage: roots = set(basis)
        sage: roots.update(-b for b in basis)
        sage: new_roots = set(roots)  # make a copy
        sage: while new_roots:
        ....:     prev_roots = new_roots
        ....:     new_roots = set()
        ....:     for a in prev_roots:
        ....:         for b in roots:
        ....:             c = a + b
        ....:             if c.quadratic_form() != 1 or c in roots:
        ....:                 continue
        ....:             new_roots.update([c, -c])
        ....:         roots.update(new_roots)
        sage: len(roots)
        240

    A classical construction of the Lie algebra of type `G_2` is
    the Lie algebra of all derivations of `\mathbf{O}` (as the
    automorphism group is the Lie group of type `G_2`). We verify
    that the derivations have the correct dimension::

        sage: len(O.derivations_basis())
        14

    We can construct the split octonions by taking the parameter `c = 1`::

        sage: SO = OctonionAlgebra(QQ, c=1)
        sage: SO
        Octonion algebra over Rational Field with parameters (-1, -1, 1)
        sage: i, j, k, l = SO.gens()
        sage: i^2 == j^2 == k^2 == -1
        True
        sage: l^2
        1
        sage: (i + l)^2
        0
        sage: (1 + l) * (1 + l).conjugate()
        0

    TESTS:

    Random elements work::

        sage: O = OctonionAlgebra(QQ)
        sage: O.random_element() in O
        True

    REFERENCES:

    - [Scha1996]_
    - :wikipedia:`octonion`
    - :wikipedia:`Split-octonion`
    - :wikipedia:`Hurwitz's_theorem_(composition_algebras)`
    """
    @staticmethod
    def __classcall_private__(cls, R, a=-1, b=-1, c=-1, names=('i', 'j', 'k', 'l')):
        r"""
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: O1 = OctonionAlgebra(QQ, 1)
            sage: O2 = OctonionAlgebra(QQ, QQ.one(), -1, QQ['x'](-1))
            sage: O3 = algebras.Octonion(QQ, QQ.one(), c=int(-1), names='ijkl')
            sage: O1 is O2
            True
            sage: O1 is O3
            True

            sage: OctonionAlgebra(ExteriorAlgebra(QQ, 3))
            Traceback (most recent call last):
            ...
            ValueError: the base ring must be a commutative ring
            sage: OctonionAlgebra(GF(2)['x','y','z'])
            Traceback (most recent call last):
            ...
            ValueError: the characteristic must not be 2
        """
        if R not in Rings().Commutative():
            raise ValueError("the base ring must be a commutative ring")
        if R.one() + R.one() == R.zero():
            raise ValueError("the characteristic must not be 2")
        a = R(a)
        b = R(b)
        c = R(c)
        names = normalize_names(4, names)
        return super().__classcall__(cls, R, a, b, c, names)

    def __init__(self, R, a, b, c, names):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ)
            sage: TestSuite(O).run()

            sage: O = OctonionAlgebra(QQ, 1, 3, 7)
            sage: TestSuite(O).run()

            sage: O = OctonionAlgebra(GF(3), 1, 2, 2)
            sage: TestSuite(O).run()

            sage: O = OctonionAlgebra(Zmod(6), -1, 2, 3)
            sage: TestSuite(O).run()

            sage: R.<a, b, c> = QQ[]
            sage: O = OctonionAlgebra(R, a, b, c)
            sage: TestSuite(O).run()
        """
        self._params = (a, b, c)
        self._module = FreeModule(R, 8)
        cat = MagmaticAlgebras(R.category()).Unital().WithBasis().FiniteDimensional()
        if a == b == c == -1:
            self.Element = Octonion
            if R in MetricSpaces():
                cat &= MetricSpaces()
        Parent.__init__(self, base=R, category=cat, names=names)

        # setup the multiplication table
        d = a * b
        e = a * c
        f = b * c
        g = a * b * c
        self._mult_table = (
          ((0, 1), (1, 1), (2, 1), (3, 1), (4, 1), (5, 1), (6, 1), (7, 1)),
          ((1, 1), (0, a), (3, -1), (2, -a), (5, -1), (4, -a), (7, 1), (6, a)),
          ((2, 1), (3, 1), (0, b), (1, b), (6, -1), (7, -1), (4, -b), (5, -b)),
          ((3, 1), (2, a), (1, -b), (0, -d), (7, -1), (6, -a), (5, b), (4, d)),
          ((4, 1), (5, 1), (6, 1), (7, 1), (0, c), (1, c), (2, c), (3, c)),
          ((5, 1), (4, a), (7, 1), (6, a), (1, -c), (0, -e), (3, -c), (2, -e)),
          ((6, 1), (7, -1), (4, b), (5, -b), (2, -c), (3, c), (0, -f), (1, f)),
          ((7, 1), (6, -a), (5, b), (4, -d), (3, -c), (2, e), (1, -f), (0, g)),
        )

    def _test_alternative(self, **options):
        r"""
        Test that ``self`` is an alternative algebra.

        An algebra `A` is *alternative* if for all `x, y \in A`, we have
        `(x * x) * y = x * (x * y)` and `(y * x) * x = y * (x * x)`.

        EXAMPLES::

            sage: R.<a,b,c> = QQ[]
            sage: O = OctonionAlgebra(R, a, b, c)
            sage: O._test_alternative()
        """
        tester = self._tester(**options)
        S = tester.some_elements()
        from sage.misc.misc import some_tuples
        for x, y in some_tuples(S, 2, tester._max_runs):
            tester.assertEqual((x * x) * y, x * (x * y))
            tester.assertEqual(y * (x * x), (y * x) * x)

    def _test_hurwitz(self, **options):
        r"""
        Test that ``self`` is an Hurwitz algebra.

        An algebra `A` is *Hurwitz* if there exists a nondegenerate quadratic
        form `N` such that `N(x y) = N(x) N(y)` for all `x, y \in A`.

        EXAMPLES::

            sage: R.<a,b,c> = QQ[]
            sage: O = OctonionAlgebra(R, a, b, c)
            sage: O._test_hurwitz()
        """
        tester = self._tester(**options)
        S = tester.some_elements()
        from sage.misc.misc import some_tuples
        for x, y in some_tuples(S, 2, tester._max_runs):
            tester.assertEqual((x * y).quadratic_form(), x.quadratic_form() * y.quadratic_form())

    def _repr_term(self, m) -> str:
        r"""
        Return a string representation of the term indexed by ``m``.

        EXAMPLES::

            sage: O = OctonionAlgebra(QQ)
            sage: [O._repr_term(i) for i in range(8)]
            ['1', 'i', 'j', 'k', 'l', 'li', 'lj', 'lk']
        """
        data = ['1', 'i', 'j', 'k', 'l', 'li', 'lj', 'lk']
        return data[m]

    def _repr_(self) -> str:
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: OctonionAlgebra(QQ)
            Octonion algebra over Rational Field
            sage: OctonionAlgebra(QQ, 1, 3, 7)
            Octonion algebra over Rational Field with parameters (1, 3, 7)
        """
        ret = f"Octonion algebra over {self.base_ring()}"
        a, b, c = self._params
        if a != -1 or b != -1 or c != -1:
            ret += f" with parameters ({a}, {b}, {c})"
        return ret

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: latex(OctonionAlgebra(QQ))
            \Bold{O}\left(\Bold{Q}\right)
            sage: latex(OctonionAlgebra(QQ, 1, 3, 7))
            \Bold{O}_{1, 3, 7}\left(\Bold{Q}\right)
        """
        from sage.misc.latex import latex
        ret = r"\Bold{O}"
        a, b, c = self._params
        if a != -1 or b != -1 or c != -1:
            ret += r"_{{{}, {}, {}}}".format(latex(a), latex(b), latex(c))
        return ret + r"\left({}\right)".format(latex(self.base_ring()))

    def _element_constructor_(self, x):
        r"""
        Construct an element of ``self``.

        EXAMPLES::

            sage: O137 = OctonionAlgebra(QQ, 1, 3, 7)
            sage: x = O137([1, 5, 2, 3, 6, -2, 3, -3]); x
            1 + 5*i + 2*j + 3*k + 6*l - 2*li + 3*lj - 3*lk
            sage: O = OctonionAlgebra(ZZ)
            sage: y = O([1, 5, 2, 3, 6, -2, 3, -3])
            sage: O(x)
            1 + 5*i + 2*j + 3*k + 6*l - 2*li + 3*lj - 3*lk
            sage: O137(y)
            1 + 5*i + 2*j + 3*k + 6*l - 2*li + 3*lj - 3*lk

            sage: xp = O137([1, 5, 2, 3, 6, -2/3, 3, -3]); xp
            1 + 5*i + 2*j + 3*k + 6*l - 2/3*li + 3*lj - 3*lk
            sage: O(xp)
            Traceback (most recent call last):
            ...
            TypeError: no conversion of this rational to integer
            sage: O([1, 5, 2, 3, 6, -2, 3, -3/4])
            Traceback (most recent call last):
            ...
            TypeError: no conversion of this rational to integer
        """
        if isinstance(x, Octonion_generic):
            v = self._module((<Octonion_generic> x).vec)
        elif self.base_ring().has_coerce_map_from(parent(x)):
            R = self.base_ring()
            v = self._module([R(x)] + [R.zero()]*7)
        else:
            v = self._module(x)
        return self.element_class(self, v)

    def some_elements(self):
        r"""
        Return some elements of ``self``.

        EXAMPLES::

            sage: O = OctonionAlgebra(ZZ)
            sage: O.some_elements()
            [2, 1, i, j, k, l, li, lj, lk,
             2 + 3*i + 4*j + 5*k + 6*l + 7*li + 8*lj + 9*lk,
             -2*j + 3*k - li - lj,
             8 - 7*i + 2*j + 13*k - 18*l + 45*li - 40*lj + 5*lk]
            sage: O = OctonionAlgebra(Zmod(6))
            sage: O.some_elements()
            [2, 1, i, j, k, l, li, lj, lk,
             2 + 3*i + 4*j + 5*k + li + 2*lj + 3*lk,
             4*j + 3*k + 5*li + 5*lj,
             2 + 5*i + 2*j + k + 3*li + 2*lj + 5*lk]
        """
        elts = [self.an_element()]
        elts.extend(self.basis())
        elt = sum(i * b for i, b in enumerate(self.basis(), start=2))
        elts.append(elt)
        elt2 = self([0, 0, -2, 3, 0, -1, -1, 0])
        elts.append(elt2)
        elts.append(elt * elt2)
        return elts

    @cached_method
    def one_basis(self):
        r"""
        Return the index for the basis element of `1`.

        EXAMPLES::

            sage: O = OctonionAlgebra(ZZ)
            sage: O.one_basis()
            0
        """
        return 0

    @cached_method
    def gens(self) -> tuple:
        r"""
        Return the generators of ``self``.

        EXAMPLES::

            sage: O = OctonionAlgebra(ZZ)
            sage: O.gens()
            (i, j, k, l)
        """
        e = self._module.basis()
        return tuple([self.element_class(self, v) for v in e[1:5]])

    @cached_method
    def basis(self):
        r"""
        Return the basis of ``self``.

        EXAMPLES::

            sage: O = OctonionAlgebra(ZZ)
            sage: O.basis()
            Family (1, i, j, k, l, li, lj, lk)
        """
        e = self._module.basis()
        from sage.sets.family import Family
        return Family([self.element_class(self, v) for v in e])

    Element = Octonion_generic



"""
Combinatorial species

This file defines the main classes for working with combinatorial
species, operations on them, as well as some implementations of
basic species required for other constructions.

This code is based on the work of Ralf Hemmecke and Martin Rubey's
Aldor-Combinat, which can be found at
http://www.risc.uni-linz.ac.at/people/hemmecke/aldor/combinat/index.html.
In particular, the relevant section for this file can be found at
http://www.risc.uni-linz.ac.at/people/hemmecke/AldorCombinat/combinatse8.html.

Weighted Species:

As a first application of weighted species, we count unlabeled
ordered trees by total number of nodes and number of internal
nodes. To achieve this, we assign a weight of `1` to the
leaves and of `q` to internal nodes::

    sage: q = QQ['q'].gen()
    sage: leaf = species.SingletonSpecies()
    doctest:warning...
    DeprecationWarning: combinat.species is superseded by LazyCombinatorialSpecies
    See https://github.com/sagemath/sage/issues/38544 for details.
    sage: internal_node = species.SingletonSpecies(weight=q)
    sage: L = species.LinearOrderSpecies(min=1)
    sage: T = species.CombinatorialSpecies(min=1)
    sage: T.define(leaf + internal_node*L(T))
    sage: T.isotype_generating_series()[0:6]                                            # needs sage.modules
    [0, 1, q, q^2 + q, q^3 + 3*q^2 + q, q^4 + 6*q^3 + 6*q^2 + q]

Consider the following::

    sage: T.isotype_generating_series().coefficient(4)                                  # needs sage.modules
    q^3 + 3*q^2 + q

This means that, among the trees on `4` nodes, one has a
single internal node, three have two internal nodes, and one has
three internal nodes.
"""
# ****************************************************************************
#       Copyright (C) 2008 Mike Hansen <mhansen@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from .generating_series import OrdinaryGeneratingSeriesRing, ExponentialGeneratingSeriesRing, CycleIndexSeriesRing
from sage.rings.rational_field import QQ
from sage.structure.sage_object import SageObject
from sage.misc.cachefunc import cached_method
from sage.combinat.species.misc import accept_size
from sage.combinat.species.structure import StructuresWrapper, IsotypesWrapper
from functools import reduce


class GenericCombinatorialSpecies(SageObject):
    def __init__(self, min=None, max=None, weight=None):
        """
        TESTS::

            sage: P = species.PermutationSpecies(size=3)
            sage: P._weight
            1
            sage: P._min
            3
            sage: P._max
            4
        """
        self._weight = weight if weight is not None else QQ(1)
        self._min = min
        self._max = max

    def __hash__(self):
        """
        Return a hash of the unique info tuple.

        EXAMPLES::

            sage: hash(species.SetSpecies()) #random
            -152204909943771174
        """
        return hash(self._unique_info())

    def _unique_info(self):
        """
        Return a tuple which should uniquely identify the species.

        EXAMPLES::

            sage: species.SetSpecies()._unique_info()
            (<class 'sage.combinat.species.set_species.SetSpecies'>, None, None, 1)
            sage: species.SingletonSpecies()._unique_info()
            (<class 'sage.combinat.species.characteristic_species.SingletonSpecies'>,
             None,
             None,
             1)

        ::

            sage: X = species.SingletonSpecies()
            sage: Y = X + X
            sage: Y._unique_info()
            (<class 'sage.combinat.species.sum_species.SumSpecies'>,
             None,
             None,
             1,
             Singleton species,
             Singleton species)
        """
        info = (self.__class__, self._min, self._max, self._weight)
        if hasattr(self, "_state_info") and self._state_info:
            info += tuple(self._state_info)
        return info

    def __eq__(self, x):
        """
        Test equality between two species.

        EXAMPLES::

            sage: X = species.SingletonSpecies()
            sage: X + X == X + X
            True
            sage: X == X
            True
            sage: X == species.EmptySetSpecies()
            False
            sage: X == X*X
            False

        ::

            sage: X = species.SingletonSpecies()
            sage: E = species.EmptySetSpecies()
            sage: L = CombinatorialSpecies()
            doctest:warning...
            DeprecationWarning: combinat.species is superseded by LazyCombinatorialSpecies
            See https://github.com/sagemath/sage/issues/38544 for details.
            sage: L.define(E+X*L)
            sage: K = CombinatorialSpecies()
            sage: K.define(E+X*L)
            sage: L == K
            True
        """
        if not isinstance(x, GenericCombinatorialSpecies):
            return False
        return self._unique_info() == x._unique_info()

    def __ne__(self, other):
        """
        Check whether ``self`` and ``other`` are not equal.

        EXAMPLES::

            sage: X = species.SingletonSpecies()
            sage: X + X == X + X
            True
            sage: X != X
            False
            sage: X != species.EmptySetSpecies()
            True
            sage: X != X*X
            True

            sage: X = species.SingletonSpecies()
            sage: E = species.EmptySetSpecies()
            sage: L = CombinatorialSpecies()
            sage: L.define(E+X*L)
            sage: K = CombinatorialSpecies()
            sage: K.define(E+X*L)
            sage: L != K
            False
        """
        return not (self == other)

    def _getstate_(self):
        r"""
        This is used during the pickling process and returns a dictionary
        of the data needed to create this object during the unpickling
        process. It returns an (\*args, \*\*kwds) tuple which is to be
        passed into the constructor for the class of this species. Any
        subclass should define a ``_state_info`` list for any arguments which
        need to be passed in the constructor.

        EXAMPLES::

            sage: C = species.CharacteristicSpecies(5)
            sage: args, kwds = C._getstate_()
            sage: args
            {0: 5}
            sage: sorted(kwds.items())
            [('max', None), ('min', None), ('weight', 1)]
        """
        kwds = {'weight': self._weight, 'min': self._min, 'max': self._max}
        try:
            return (dict(enumerate(self._state_info)), kwds)
        except AttributeError:
            return ({}, kwds)

    def _setstate_(self, state):
        """
        This is used during unpickling to recreate this object from the
        data provided by the ``_getstate_`` method.

        TESTS::

            sage: C2 = species.CharacteristicSpecies(2)
            sage: C4 = species.CharacteristicSpecies(4)
            sage: C2
            Characteristic species of order 2
            sage: C2._setstate_(C4._getstate_()); C2
            Characteristic species of order 4
        """
        args_dict, kwds = state
        self.__class__.__init__(self, *[args_dict[i] for i in range(len(args_dict))], **kwds)

    def weighted(self, weight):
        """
        Return a version of this species with the specified weight.

        EXAMPLES::

            sage: t = ZZ['t'].gen()
            sage: C = species.CycleSpecies(); C
            Cyclic permutation species
            sage: C.weighted(t)
            Cyclic permutation species with weight=t
        """
        args_dict, kwds = self._getstate_()
        kwds.update({'weight': weight})
        return self.__class__(*[args_dict[i] for i in range(len(args_dict))], **kwds)

    def __repr__(self):
        """
        Return a string representation of this species.

        EXAMPLES::

            sage: CombinatorialSpecies()
            Combinatorial species

        ::

            sage: species.SetSpecies()
            Set species
            sage: species.SetSpecies(min=1)
            Set species with min=1
            sage: species.SetSpecies(min=1, max=4)
            Set species with min=1, max=4
            sage: t = ZZ['t'].gen()
            sage: species.SetSpecies(min=1, max=4, weight=t)
            Set species with min=1, max=4, weight=t
        """
        if hasattr(self, "_name"):
            name = self._name if isinstance(self._name, str) else self._name()
        else:
            name = "Combinatorial species"

        options = []

        if self._min is not None:
            options.append('min=%s' % self._min)
        if self._max is not None:
            options.append('max=%s' % self._max)
        if self._weight != 1:
            options.append('weight=%s' % self._weight)

        if options:
            name += " with " + ", ".join(options)

        return name

    def __add__(self, g):
        """
        Return the sum of ``self`` and ``g``.

        EXAMPLES::

            sage: P = species.PermutationSpecies()
            sage: F = P + P; F
            Sum of (Permutation species) and (Permutation species)
            sage: F.structures([1,2]).list()
            [[1, 2], [2, 1], [1, 2], [2, 1]]
        """
        from .sum_species import SumSpecies
        if not isinstance(g, GenericCombinatorialSpecies):
            raise TypeError("g must be a combinatorial species")
        return SumSpecies(self, g)

    sum = __add__

    def __mul__(self, g):
        """
        Return the product of ``self`` and ``g``.

        EXAMPLES::

            sage: P = species.PermutationSpecies()
            sage: F = P * P; F
            Product of (Permutation species) and (Permutation species)
        """
        from .product_species import ProductSpecies
        if not isinstance(g, GenericCombinatorialSpecies):
            raise TypeError("g must be a combinatorial species")
        return ProductSpecies(self, g)

    product = __mul__

    def __call__(self, g):
        """
        EXAMPLES::

            sage: S = species.SetSpecies()
            sage: S(S)
            Composition of (Set species) and (Set species)
        """
        from .composition_species import CompositionSpecies
        if not isinstance(g, GenericCombinatorialSpecies):
            raise TypeError("g must be a combinatorial species")
        return CompositionSpecies(self, g)

    composition = __call__

    def functorial_composition(self, g):
        """
        Return the functorial composition of ``self`` with ``g``.

        EXAMPLES::

            sage: E = species.SetSpecies()
            sage: E2 = E.restricted(min=2, max=3)
            sage: WP = species.SubsetSpecies()
            sage: P2 = E2*E
            sage: G = WP.functorial_composition(P2)
            sage: G.isotype_generating_series()[0:5]                                    # needs sage.modules
            [1, 1, 2, 4, 11]
        """
        from .functorial_composition_species import FunctorialCompositionSpecies
        if not isinstance(g, GenericCombinatorialSpecies):
            raise TypeError("g must be a combinatorial species")
        return FunctorialCompositionSpecies(self, g)

    @accept_size
    def restricted(self, min=None, max=None):
        """
        Return the restriction of the species.

        INPUT:

        - ``min`` -- (optional) integer

        - ``max`` -- (optional) integer

        EXAMPLES::

            sage: S = species.SetSpecies().restricted(min=3); S
            Set species with min=3
            sage: S.structures([1,2]).list()
            []
            sage: S.generating_series()[0:5]
            [0, 0, 0, 1/6, 1/24]
        """
        kwargs = {'min': self._min if min is None else min,
                  'max': self._max if max is None else max,
                  'weight': self._weight}
        return self.__class__(**kwargs)

    def structures(self, labels, structure_class=None):
        """
        EXAMPLES::

            sage: F = CombinatorialSpecies()
            sage: F.structures([1,2,3]).list()
            Traceback (most recent call last):
            ...
            ValueError: stream is not yet defined
        """
        return StructuresWrapper(self, labels, structure_class)

    def isotypes(self, labels, structure_class=None):
        """
        EXAMPLES::

            sage: F = CombinatorialSpecies()
            sage: F.isotypes([1,2,3]).list()
            Traceback (most recent call last):
            ...
            ValueError: stream is not yet defined
        """
        return IsotypesWrapper(self, labels, structure_class=structure_class)

    def _check(self, n=5):
        """
        Return ``True`` if the number of structures and isomorphism types
        generated is the same as the number found from the generating
        series.

        EXAMPLES::

            sage: P = species.PartitionSpecies()
            sage: P._check()                                                            # needs sage.libs.flint
            True
        """
        st = self.structures(range(n))
        it = self.isotypes(range(n))

        try:
            return (len(st.list()) == st.cardinality() and
                    len(it.list()) == it.cardinality())
        except NotImplementedError:
            return False

    def __pow__(self, n):
        r"""
        Return this species to the power `n`.

        This uses a binary exponentiation algorithm to perform the
        powering.

        EXAMPLES::

            sage: One = species.EmptySetSpecies()
            sage: X = species.SingletonSpecies()
            sage: X^2
            Product of (Singleton species) and (Singleton species)
            sage: X^5
            Product of (Singleton species) and (Product of (Product of
            (Singleton species) and (Singleton species)) and (Product
            of (Singleton species) and (Singleton species)))

            sage: (X^2).generating_series()[0:4]
            [0, 0, 1, 0]
            sage: (X^3).generating_series()[0:4]
            [0, 0, 0, 1]
            sage: ((One+X)^3).generating_series()[0:4]
            [1, 3, 3, 1]
            sage: ((One+X)^7).generating_series()[0:8]
            [1, 7, 21, 35, 35, 21, 7, 1]

            sage: x = QQ[['x']].gen()
            sage: coeffs = ((1+x+x+x**2)**25+O(x**10)).padded_list()
            sage: T = ((One+X+X+X^2)^25)
            sage: T.generating_series()[0:10] == coeffs
            True
            sage: X^1 is X
            True
            sage: A = X^32
            sage: A.digraph()                                                           # needs sage.graphs
            Multi-digraph on 6 vertices

        TESTS::

            sage: X**(-1)
            Traceback (most recent call last):
            ...
            ValueError: only positive exponents are currently supported
        """
        from sage.rings.integer import Integer
        import operator
        n = Integer(n)
        if n <= 0:
            raise ValueError("only positive exponents are currently supported")
        digits = n.digits(2)
        squares = [self]
        for i in range(len(digits) - 1):
            squares.append(squares[-1] * squares[-1])
        return reduce(operator.mul, (s for i, s in zip(digits, squares)
                                     if i != 0))

    def _get_series(self, series_ring_class, prefix, base_ring=None):
        """
        Return the generating / isotype generating / cycle index series
        ring. The purpose of this method is to restrict the result of
        _series_helper to ``self._min`` and ``self._max``.

        EXAMPLES::

            sage: P = species.PermutationSpecies(min=2, max=4)
            sage: P.generating_series()[0:8] #indirect doctest
            [0, 0, 1, 1, 0, 0, 0, 0]
        """
        series = self._series_helper(series_ring_class, prefix, base_ring=base_ring)
        # We need to restrict the series based on the min
        # and max of this species.  Note that if min and max
        # are both None (as in the default case), then the restrict
        # method will just return series.
        if self._min is None and self._max is None:
            return series
        return series.parent()(lambda n: series[n],
                               valuation=self._min, degree=self._max)

    def _series_helper(self, series_ring_class, prefix, base_ring=None):
        """
        This code handles much of the common work involved in getting the
        generating series for this species (such as determining the
        correct base ring to pass down to the subclass, determining which
        method on the subclass to call to get the series object, etc.)

        INPUT:

        - ``series_ring_class`` -- a class for the series ring such as
          ``ExponentialGeneratingSeriesRing``, etc.

        - ``prefix`` -- the string prefix associated with the generating series
          such as "cis" for the cycle index series. This prefix appears in the
          methods that are implemented in the subclass.

        - ``base_ring`` -- the ring in which the coefficients of the generating
          series live. If it is not specified, then it is determined by the
          weight of the species.

        EXAMPLES::

            sage: from sage.combinat.species.generating_series import OrdinaryGeneratingSeriesRing
            sage: S = species.SetSpecies()
            sage: itgs = S._series_helper(OrdinaryGeneratingSeriesRing, "itgs")
            sage: itgs[:3]
            [1, 1, 1]

        ::

            sage: itgs = S._series_helper(OrdinaryGeneratingSeriesRing, "itgs", base_ring=RDF)
            sage: itgs[:3]
            [1.0, 1.0, 1.0]
        """
        prefix = "_" + prefix

        # Get the base ring
        if base_ring is None:
            base_ring = self.weight_ring()
        else:
            # The specified base ring must have maps from both
            # the rational numbers and the weight ring
            if not base_ring.has_coerce_map_from(QQ):
                raise ValueError("specified base ring does not contain the rationals")
            if not base_ring.has_coerce_map_from(self.weight_ring()):
                raise ValueError("specified base ring is incompatible with the weight ring of self")

        series_ring = series_ring_class(base_ring)

        # Try to return things like self._gs(base_ring)
        # This is used when the subclass wants to just
        # handle creating the generating series itself;
        # for example, returning the exponential of a
        # generating series.
        try:
            return getattr(self, prefix)(series_ring, base_ring)
        except AttributeError:
            pass

        # Try to return things like self._gs_callable(base_ring).
        # This is used when the subclass just provides a callable
        # for the coefficients of the generating series.  Optionally,
        # the subclass can specify the order of the series.
        try:
            callable = getattr(self, prefix + "_callable")
            try:
                return series_ring(lambda n: callable(base_ring, n), valuation=self._order())
            except AttributeError:
                return series_ring(lambda n: callable(base_ring, n))
        except AttributeError:
            pass

        # Try to use things like self._gs_term(base_ring).
        # This is used when the generating series is just a single
        # term.
        try:
            return series_ring(getattr(self, prefix + "_term")(base_ring),
                                    self._order())
        except AttributeError:
            pass

        # Try to use things like self._gs_list(base_ring).
        # This is used when the coefficients of the generating series
        # can be given by a finite list with the last coefficient repeating.
        # The generating series with all ones coefficients is generated this
        # way.
        try:
            return series_ring(lambda n: getattr(self, prefix + "_list")(base_ring, n))
        except AttributeError:
            pass

        raise NotImplementedError

    @cached_method
    def generating_series(self, base_ring=None):
        r"""
        Return the generating series for this species.

        This is an exponential generating series so the `n`-th
        coefficient of the series corresponds to the number of labeled
        structures with `n` labels divided by `n!`.

        EXAMPLES::

            sage: P = species.PermutationSpecies()
            sage: g = P.generating_series()
            sage: g[:4]
            [1, 1, 1, 1]
            sage: g.counts(4)
            [1, 1, 2, 6]
            sage: P.structures([1,2,3]).list()
            [[1, 2, 3], [1, 3, 2], [2, 1, 3], [2, 3, 1], [3, 1, 2], [3, 2, 1]]
            sage: len(_)
            6
        """
        return self._get_series(ExponentialGeneratingSeriesRing, "gs", base_ring)

    @cached_method
    def isotype_generating_series(self, base_ring=None):
        r"""
        Return the isotype generating series for this species.

        The `n`-th coefficient of this series corresponds to the number
        of isomorphism types for the structures on `n` labels.

        EXAMPLES::

            sage: P = species.PermutationSpecies()
            sage: g = P.isotype_generating_series()
            sage: g[0:4]                                                                # needs sage.libs.flint
            [1, 1, 2, 3]
            sage: g.counts(4)                                                           # needs sage.libs.flint
            [1, 1, 2, 3]
            sage: P.isotypes([1,2,3]).list()                                            # needs sage.libs.flint
            [[2, 3, 1], [2, 1, 3], [1, 2, 3]]
            sage: len(_)                                                                # needs sage.libs.flint
            3
        """
        return self._get_series(OrdinaryGeneratingSeriesRing, "itgs", base_ring)

    @cached_method
    def cycle_index_series(self, base_ring=None):
        r"""
        Return the cycle index series for this species.

        The cycle index series is a sequence of symmetric functions.

        EXAMPLES::

            sage: P = species.PermutationSpecies()
            sage: g = P.cycle_index_series()                                            # needs sage.modules
            sage: g[0:4]                                                                # needs sage.modules
            [p[], p[1], p[1, 1] + p[2], p[1, 1, 1] + p[2, 1] + p[3]]
        """
        return self._get_series(CycleIndexSeriesRing, "cis", base_ring)

    def is_weighted(self):
        """
        Return ``True`` if this species has a nontrivial weighting associated
        with it.

        EXAMPLES::

            sage: C = species.CycleSpecies()
            sage: C.is_weighted()
            False
        """
        return self._weight != 1

    def weight_ring(self):
        """
        Return the ring in which the weights of this species occur.

        By default, this is just the field of rational numbers.

        EXAMPLES::

            sage: species.SetSpecies().weight_ring()
            Rational Field
        """
        if self.is_weighted():
            return self._weight.parent()
        return QQ

    def _common_parent(self, parents):
        """
        Return a parent that contains all the parents
        in the given list of parents.

        EXAMPLES::

            sage: C = species.CombinatorialSpecies()
            sage: C._common_parent([QQ, ZZ['t']])
            Univariate Polynomial Ring in t over Rational Field
        """
        assert parents
        from sage.structure.element import get_coercion_model
        cm = get_coercion_model()

        common = parents[0]
        for p in parents[1:]:
            common = cm.explain(common, p, verbosity=0)
            if common is None:
                raise ValueError("unable to find a common parent")
        return common

    def digraph(self):
        """
        Return a directed graph where the vertices are the individual
        species that make up this one.

        EXAMPLES::

            sage: X = species.SingletonSpecies()
            sage: B = species.CombinatorialSpecies()
            sage: B.define(X+B*B)
            sage: g = B.digraph(); g                                                    # needs sage.graphs
            Multi-digraph on 4 vertices

            sage: sorted(g, key=str)                                                    # needs sage.graphs
            [Combinatorial species,
             Product of (Combinatorial species) and (Combinatorial species),
             Singleton species,
             Sum of (Singleton species) and
              (Product of (Combinatorial species) and (Combinatorial species))]

            sage: d = {sp: i for i, sp in enumerate(g)}                                 # needs sage.graphs
            sage: g.relabel(d)                                                          # needs sage.graphs
            sage: g.canonical_label().edges(sort=True)                                  # needs sage.graphs
            [(0, 3, None), (2, 0, None), (2, 0, None), (3, 1, None), (3, 2, None)]
        """
        from sage.graphs.digraph import DiGraph
        d = DiGraph(multiedges=True)
        self._add_to_digraph(d)
        return d

    def _add_to_digraph(self, d):
        """
        Add this species as a vertex to the digraph ``d`` along with any
        'children' of this species. For example, sum species would add
        itself as a vertex and an edge between itself and each of its
        summands.

        EXAMPLES::

            sage: # needs sage.graphs
            sage: d = DiGraph(multiedges=True)
            sage: X = species.SingletonSpecies()
            sage: X._add_to_digraph(d); d
            Multi-digraph on 1 vertex
            sage: (X+X)._add_to_digraph(d); d
            Multi-digraph on 2 vertices
            sage: d.edges(sort=True)
            [(Sum of (Singleton species) and (Singleton species), Singleton species, None),
             (Sum of (Singleton species) and (Singleton species), Singleton species, None)]
        """
        d.add_vertex(self)

        if not hasattr(self, "_state_info"):
            return

        for child in self._state_info:
            if not isinstance(child, GenericCombinatorialSpecies):
                continue
            d.add_edge(self, child)
            child._add_to_digraph(d)

    def algebraic_equation_system(self):
        """
        Return a system of algebraic equations satisfied by this species.

        The nodes are numbered in the order that they appear as vertices of
        the associated digraph.

        EXAMPLES::

            sage: B = species.BinaryTreeSpecies()
            sage: B.algebraic_equation_system()                                         # needs sage.graphs
            [-node3^2 + node1, -node1 + node3 + (-z)]

        ::

            sage: sorted(B.digraph().vertex_iterator(), key=str)                        # needs sage.graphs
            [Combinatorial species with min=1,
             Product of (Combinatorial species with min=1)
                    and (Combinatorial species with min=1),
             Singleton species,
             Sum of (Singleton species)
                and (Product of (Combinatorial species with min=1)
                and (Combinatorial species with min=1))]

        ::

            sage: B.algebraic_equation_system()[0].parent()                             # needs sage.graphs
            Multivariate Polynomial Ring in node0, node1, node2, node3 over
             Fraction Field of Univariate Polynomial Ring in z over Rational Field
        """
        d = self.digraph()

        Qz = QQ['z'].fraction_field()

        # Generate the variable names and the corresponding polynomial rings
        var_names = ["node%s" % i for i in range(d.n_vertices())]
        R = Qz[", ".join(var_names)]
        R_gens_dict = R.gens_dict()

        # A dictionary mapping the nodes to variables
        vertices = sorted(d.vertex_iterator(), key=str)
        var_mapping = {node: R_gens_dict[name]
                       for node, name in zip(vertices, var_names)}
        var_mapping['z'] = Qz.gen()

        eqns = []
        subs = {}
        for species in vertices:
            try:
                eqn = species._equation(var_mapping)
                if eqn in Qz or eqn in R.gens():
                    subs[var_mapping[species]] = eqn
                else:
                    eqns.append(var_mapping[species] - eqn)
            except AttributeError:
                raise NotImplementedError
        return [eq.subs(subs) for eq in eqns]



from sage.libs.flint.acb cimport acb_t
from sage.rings.complex_interval cimport ComplexIntervalFieldElement
from sage.rings.real_arb cimport RealBall
from sage.structure.element cimport RingElement
from sage.rings.ring cimport Field

cdef void ComplexIntervalFieldElement_to_acb(
    acb_t target,
    ComplexIntervalFieldElement source) noexcept

cdef int acb_to_ComplexIntervalFieldElement(
    ComplexIntervalFieldElement target,
    const acb_t source) except -1

cdef class ComplexBall(RingElement):
    cdef acb_t value
    cdef ComplexBall _new(self)
    cpdef _add_(self, other)
    cpdef _mul_(self, other)
    cpdef ComplexIntervalFieldElement _complex_mpfi_(self, parent)
    cpdef RealBall real(self)
    cpdef RealBall imag(self)
    cpdef pow(self, expo, analytic=?)

    cdef inline ComplexBall _new(self):
        cdef ComplexBall res = ComplexBall.__new__(ComplexBall)
        res._parent = self._parent
        return res
